/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

/* 
 * authentication-only crypto i/o
 *
 * Authentication-only crypto.  You know who you are talking to,
 * but the conversation is not secret.
 */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#include "crypto.h"

KEY	read_key;
KEY	write_key;

u_llong	read_seq;
u_llong	write_seq;

char protohead[1+20+2+MAXSERVICENAME];
int headlen;


static char *
err(char *msg) {
	static char buf[MAXIO];

	sprintf(buf, "%s: %s", msg, sslerror);
	return buf;
}

char *
proto_setup_read(int fd, void *msg, int n) {
	static char buf[200];
	int i;

	if ((i=sslread(fd, buf, sizeof(buf))) < 0) {
		if (i == C_CORRUPTED)
			return "secret keys don't match";
		else
			return err("reading protocol setup");
	}
	if (protohead[0] == buf[0] ||	/* don't feed us our own stuff */
	    i < headlen ||
	    memcmp(&protohead[1], &buf[1], headlen-1) != 0) {
		return "protocol setup message format error, or wrong protocol version";
	}

	if (strncmp(&buf[headlen], ERRSTR, strlen(ERRSTR)) == 0)
		return &buf[headlen];
	if (i - headlen > n)
		return "protocol setup message too long";

	memcpy(msg, &buf[headlen], i - headlen);
	return 0;
}

char *
proto_setup_write(int fd, void *msg, int n) {
	char buf[1000];

	if (n + headlen > sizeof(buf))
		return "proto header message too long";

	memcpy(buf, protohead, headlen);
	memcpy(&buf[headlen], msg, n);
	if (sslwrite(fd, buf, headlen+n) < 0)
		return err("writing protocol initialization");
	return 0;
}


void
make_challenge(u_char *key) {
	DIGEST_CTX context;
	u_long *lp = (u_long *)key;
	int i;

	for (i=0; i<sizeof(KEY)/4; i++)
		*lp++ = truerand();

	/* whiten the result */
	HASH_INIT(&context);
	HASH_UPDATE(&context, key, sizeof(key));
	HASH_FINAL(&context, key);
}

void
chal_to_key(u_char *chal_a, u_char *chal_b, u_char *master, u_char *key) {
	hmac_ctx hmac_context;

	hmac_init(&hmac_context, master, sizeof(KEY));
	hmac_update(&hmac_context, chal_a, sizeof(KEY));
	hmac_update(&hmac_context, chal_b, sizeof(KEY));
	hmac_final(&hmac_context, key);
}


/*
 * read a static key from a file. (yawn)
 */
char *
readkey(char *fn, u_char *key) {
	static char buf[MAXIO];
	int fd = open(fn, 0);
	int n;

	if (fd < 0) {
		sprintf(buf, "could not open key file '%s': %s",
			fn, strerror(errno));
		return buf;
	}

	n = read(fd, buf, sizeof(buf));
	if (n < 0) {
		sprintf(buf, "error reading key file '%s': %s",
			fn, strerror(errno));
		return buf;
	}
	close(fd);

	memset(key, 0, sizeof(key));
	/*
	 * Check the key format.   "0x" means a sequence of bytes.
	 * If the key is longer than our standard key, xor
	 * the extra bytes.  The key may not be too short.
	 */
	if (strncmp(buf, "0x", 2) == 0) {
		int n=0;
		char *cp = buf;
		u_long byte;

		do {
			cp += 2;
			byte = strtoul(cp, &cp, 16);
			if (byte > 255)
				goto bad;
			if (n < sizeof(KEY))
				key[n] = byte;
			else
				key[n % sizeof(KEY)] ^= byte;
			n++;
			cp += strspn(cp, ", \t\n");
		} while (strncmp(cp, "0x", 2) == 0);
		if (n < sizeof(KEY)) {
			sprintf(buf,
				"key file '%s': key is too short\n",
				fn);
			return buf;
		}
		return 0;
	}

	/*
	 * Try 64-encoding.
	 */
	
bad:
	sprintf(buf, "key file '%s': bad key format\n", fn);
	return buf;
}
