/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <string.h>

#include "crypto.h"
#include "util.h"

/*
 * Set up an authenticated session with the caller.
 */
char *
start_server_crypto(int fd) {
	char msg[20+MAXSERVICENAME+1+sizeof(KEY)];
	static char errbuf[200];
	int n;
	KEY c_chal, s_chal;
	KEY key;
	char *cp, *ep, *servicename;

	memset(write_key, 0, sizeof(write_key));
	memset(read_key, 0, sizeof(read_key));
	read_seq = 0;
	write_seq = 0x4000000000000000LL;

	make_challenge(s_chal);

	sprintf(protohead, "%c%d:", 'S', PROTOVER);
	headlen = strlen(protohead)+1;

	n = sslread(fd, msg, sizeof(msg));
	if (n <= 0) {
		sprintf(errbuf, "%s initial read error - %s",
			ERRSTR, sslerror);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return &errbuf[sizeof(ERRSTR)];
	}
	if (n < sizeof("C1:") ||
	    msg[0] != 'C' ||
	    strtol(&msg[1], &cp, 10) != PROTOVER ||
	    *cp++ != ':') {
		sprintf(errbuf, "%s - illegal initial protocol message", ERRSTR);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return &errbuf[sizeof(ERRSTR)];
	}
	servicename = cp = &msg[strlen(msg) + 1];

	if (strchr(servicename, '/') != 0 || strlen(servicename) > MAXSERVICENAME) {
		sprintf(errbuf, "%s illegal service name", ERRSTR);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return &errbuf[sizeof(ERRSTR)];
	}

	if ((ep=setservice(servicename, key)) != 0) {
		sprintf(errbuf, "%s %s: %s", ERRSTR, ep, servicename);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return &errbuf[sizeof(ERRSTR)];
	}
	cp += strlen(servicename) + 1;
	memcpy(c_chal, cp, sizeof(KEY));

	if ((cp=proto_setup_write(fd, s_chal, sizeof(s_chal))) != 0)
		return cp;

	strcat(protohead, servicename);
	headlen = strlen(protohead) + 1;

	/*
	 * Compute keys, and prove we know them.
	 */
	chal_to_key(c_chal, s_chal, key, write_key);
	chal_to_key(s_chal, c_chal, key, read_key);
	memset(key, 0, sizeof(KEY));	/* clear the secret */
	if ((cp=proto_setup_read(fd, msg, 100)) != 0) {
		sprintf(errbuf, "%s %s", ERRSTR);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return cp;
	}

	if (strcmp(msg, "OK") != 0) {
		sprintf(errbuf, "%s secret keys don't match", ERRSTR);
		proto_setup_write(fd, errbuf, strlen(errbuf)+1);
		return "secret keys don't match";
	}

	if ((cp=proto_setup_write(fd, "OK", 3)) != 0)
		return cp;
	return 0;
}
