/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <stdio.h>
#include <sys/types.h>

#include "crypto.h"

/*
 * Function: hmac_sha - from Internet Draft, Krawczyk, Bellare, Canetti
 */

void
hmac_init(hmac_ctx *hmac_context, u_char *key, int key_len) {
	int i;

	if (key == 0 && key_len > 64) {
		fprintf(stderr, "hac_init: key too long: %d\n", key_len);
		exit(1);
	}

	/*
	 * the HMAC_MD5 transform looks like:
	 *
	 * MD5(K XOR opad, MD5(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times
	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected
	 */

	/* start out by storing key in pads */
	memset(hmac_context->k_ipad, 0, sizeof(hmac_context->k_ipad));
	memset(hmac_context->k_opad, 0, sizeof(hmac_context->k_opad));
	memcpy(hmac_context->k_ipad, key, key_len);
	memcpy(hmac_context->k_opad, key, key_len);
 
	/* XOR key with ipad and opad values */
	for (i=0; i<64; i++) {
		hmac_context->k_ipad[i] ^= 0x36;
		hmac_context->k_opad[i] ^= 0x5c;
	}
	/*
	 * perform inner hash
	 */
	HASH_INIT(&hmac_context->context);
	/*
	 * start with inner pad
	 */
	HASH_UPDATE(&hmac_context->context, hmac_context->k_ipad, 64);
}

void
hmac_update(hmac_ctx *hmac_context, u_char *text, int text_len) {
	HASH_UPDATE(&hmac_context->context, text, text_len);
}

void
hmac_final(hmac_ctx *hmac_context, u_char *digest) {
	DIGEST dig;
	DIGEST_CTX context;

	/*
	 * finish up 1st pass
	 */
	HASH_FINAL(&hmac_context->context, dig);
	/*
	 * perform outer hash, starting with outer pad.
	 */
	HASH_INIT(&context);
	HASH_UPDATE(&context, hmac_context->k_opad, 64);
	HASH_UPDATE(&context, dig, sizeof(dig));
	HASH_FINAL(&context, digest);
}
