/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "u.h"
#include "crypto.h"

char cerror[BUFSIZ];

char *cname = "";

void
cperror(char *msg) {
	fprintf(stderr, "%s\n", cerror);
}

int
cread(int fd, void *ibuf, int blen) {
	u_char count[2];
	u_char p[2];
	int len;

	/* XXX - endian trouble? */
	if (netread(fd, count, 2) < 0) {
		sprintf(cerror, "cread failed: %s", strerror(errno));
		return -1;
	}
	if(count[0] & 0x80){
		len = ((count[0] & 0x7f)<<8) | count[1];
	} else {
		/*
		 * Padding is present, which we don't use.
		 */
		len = ((count[0] & 0x3f)<<8) | count[1];
		fprintf(stderr, "Padding!\n");
		return -2;
	}
	if (len > blen || len > BUFSIZ) {
		fprintf(stderr, "incoming record too long, %d > %d",
			len, blen);
		return -3;
	}

	if (netread(fd, ibuf, len) < 0) {
		sprintf(cerror, "cread failed reading data: %s",
			strerror(errno));
		return -4;
	}
	if (debug > 2) {
		fprintf(stderr, "  txt: %s...",
			ds((u_char *)ibuf, 4));
		fprintf(stderr, "%s\n",
			ds((u_char *)ibuf+len-4, 4));
	}
	return len;
}

void
cwrite(int fd, void *obuf, int len) {
	int m;
	u_char p[2];

	m = len;
	if (len >= 0x8000) {
		fprintf(stderr, "cwrite, block too large, %d > %d\n",
			len, 0x8000);
		}

	p[0] = (m>>8) | 0x80;
	p[1] = m & 0xff;

	netwrite(fd, p, 2, "cwrite: count");
	netwrite(fd, obuf, len, "cwrite: text");
}
