/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <sys/param.h>
#include <string.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <errno.h>

#include "stage.h"
#include "crypto.h"
#include "util.h"

char	response[MAXIO];	/* Response lines from the server */
int 	remok;
char	remerr[MAXPATHLEN];

extern	void log(char *msg, ...);
extern	int cmd_count;
extern	int responses;
extern	int response_wait;

mode_t	srv_umask;
mode_t	srv_or_perm;

/*
 * get server mask information, so we can determine if permissions
 * need adjustment.
 */
void
get_server_info(int srvfd) {
	int n;
	char buf[MAXIO];

	remcmd(srvfd, "si", "");
	remresp(srvfd);
	if (!remok) {
		fprintf(stderr, "stage: remote needs newest version of staged server\n");
		exit(1);
	}
	n = sscanf(response, "%o %o",
		(mode_t *)&srv_umask, (mode_t *)&srv_or_perm);
	if (n != 2) {
		fprintf(stderr, "Remote server, si format error: %s\n",
			response);
		exit(1);
	}
}

/*
 * Compute the path on the remote host, given our local path and the
 * location of the root of the local directory. Path must be absolute.
 */
char *
remotepath(char *path) {
	if (strcmp(root, path) == 0)	/* it is the root */
		return "/";
	if (strlen(root) >= strlen(path) ||
	    strncmp(root, path, strlen(root)) != 0 ||
	    path[strlen(root)] != '/') {
		fprintf(stderr, "file %s is not in the local staged tree %s, or local tree root is invalid\n",
			path, root);
		exit(1);
	}
	return &path[strlen(root)];
}

/*
 * fetch status fields of a remote file.
 */
struct stat *
get_remote_stat(int fd, char *lpath) {
	remcmd(fd, "st", remotepath(lpath));
	remresp(fd);
	if (!remok)
		return 0;
	else
		return crackstat(response);
}

/*
 * Send a command to the server.
 */
void
remcmd(int fd, char *cmd, char *path) {
	char buf[MAXIO];

	sprintf(buf, "%s %s", cmd, path);
	if (vflg > 1)
		fprintf(stderr, "<--- %s\n", buf);
	cwrites(fd, buf);
	cmd_count++;
}

/*
 * Obtain a remote directory list.  We read one or more blocks until
 * a null terminator is found.
 */
char *
get_remote_ls(int fd, char *lpath) {
	char buf[MAXIO];
	char *pp = 0;
	int n;
	int next = 0;

	remcmd(fd, "ls", remotepath(lpath));

	do {
		n = cread(fd, buf, sizeof(buf));
		if (n < 0) {
			log("get_remote_ls: read error: %s", cerror);
			if (pp)
				free(pp);
			return 0;
		}
		pp = (char *)realloc(pp, next+n);
		if (pp == 0) {
			fprintf(stderr, "get_remote_ls: malloc: out of memory\n");
			exit(13);
		}
		memcpy(&pp[next], buf, n);
		next += n;
	} while (buf[n-1] != '\0');
	if (vflg > 1)
		fprintf(stderr, "---> %s\n", pp);
	return pp;
}

int
remresp(int fd) {
	int start = rtime();

	responses++;
	if (cread(fd, response, sizeof(response)) < 0) {
		log("cannot read reply: %s", cerror);
		exit(6);
	}
	response_wait += (rtime() - start);
	
	if (vflg > 1)
		fprintf(stderr, "---> %s\n", response);
	if (strncmp(response, "NO", 2) == 0) {
		strncpy(remerr, &response[3], sizeof(remerr)-1);
		remok = 0;
	} else
		remok = 1;
	return remok;
}
