/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <pwd.h>

#include "stage.h"
#include "crypto.h"

char user[MAXPATHLEN], group[MAXPATHLEN];

extern	void log(char *msg, ...);

struct stat laststat;

/*
 * Crack the encoded stat information. 
 */
struct stat *
crackstat(char *in) {
	int n, mode;

	if (strncmp(in, "ENOENT", 6) == 0)
		return 0;

	n=sscanf(in, "%s %s %o %d %d",
		user, group, &mode, &laststat.st_mtime, &laststat.st_size);
	if (n != 5) {
		log("stat proto error: %s", in);
		return 0;
	}
	laststat.st_mode = mode;
	return &laststat;
}

char *
statinfo(char *path) {
	static char buf[MAXPATHLEN];
	int mode;

	if (path[0] == '\0')
		path = "/";
	if (stat(path, &laststat) < 0) {
		if (errno == ENOENT)
			return "ENOENT";
		else {
			log("stat error %s", strerror(errno));
			return strerror(errno);
		}
	}

	sprintf(buf, "%d %d %o %d %d",
		laststat.st_uid, laststat.st_gid,
		(unsigned short)laststat.st_mode,
		laststat.st_mtime, laststat.st_size);
	return buf;
}

char *
checksum(char *path) {
	static char sum[sizeof(DIGEST_CTX)*2 + 1];
	char buf[MAXIO];
	u_char d[16];
	DIGEST_CTX context;
	int i, n, fd;

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return "";

	HASH_INIT(&context);
	while ((n=read(fd, buf, sizeof(buf))) > 0)
		HASH_UPDATE(&context, (u_char *)buf, n);
	HASH_FINAL(&context, d);
	close(fd);

	sum[0] = '\0';
	for (i=0; i<sizeof(DIGEST); i++) {
		sprintf(buf, "%.2x", d[i]);
		strcat(sum, buf);
	}
	return sum;
}

#define MEMINCR	1024

/*
 * Assemble a list of file/directory names in this directory,
 * entirely delimited by slashes.
 */
char *
dirls(char *path) {
	int next=1, len=MEMINCR;
	DIR *dp;
	struct dirent *ent;
	char *pp = (char *)malloc(MEMINCR);

	if (pp == 0) {
		log("dirls: out of memory");
		exit(10);
	}

	dp = opendir(path);
	if (dp == NULL) {
		strcpy(pp, "//");
		return pp;
	}

	pp[0] = '/';
	while ((ent=readdir(dp)) != NULL) {
		int namelen = strlen(ent->d_name);
		if (strcmp(ent->d_name, ".") == 0 ||
		    strcmp(ent->d_name, "..") == 0)
			continue;
		if (next + namelen + 2 > len) {
			len += MEMINCR + namelen + 2;
			pp = (char *)realloc(pp, len);
			if (pp == NULL) {
				log("dirls: Out of memory");
				exit(10);
			}
		}
		strncpy(&pp[next], ent->d_name, namelen);
		next += namelen;
		pp[next++] = '/';
	}
	pp[next++] = '/';
	pp[next] = '\0';
	closedir(dp);
	return pp;
}
