/*-------------------------------------------------------*/
/* parallel solution of the wave equation in MPI-Java    */
/* subject to initial triangular distribution and zero-->*/
/* boundary conditions at the outer ends                 */
/* Ahmed Abdelwahab                                      */
/*----------------------------------------------------   */

import MPI.*;
public class hw5{
 static public void main(String [] args){
  MPI.Init(args);
  int i,ilocal,iglobal,nlocal,k;

  final int n=35;       /* total number of points in the domain */

  final int maxiter=100;  /*  maximum number of iterations */

  double tow;
  final double delt=0.001;   /*  time step  */
  final double pi=1.0;
  double xin[]=new double[100];
  double xout[]=new double[100];
  double xtemp[]=new double[100];
  double xtemp1[]=new double[1];
  int p=MPI.COMM_WORLD.Size();
  int myrank=MPI.COMM_WORLD.Rank();
  
  nlocal = n/p;   /* local number of points  */
 
  tow = ((n-1)*delt/pi)*((n-1)*delt/pi);

/* initialization of the initial values */
/* each processor generates its own relevant part of the initial conditions */
/* the initial values are distributed as an even triangle with zero 
   oscillations at both ends */

  for(ilocal=0;ilocal<nlocal;ilocal++){
    iglobal=ilocal+myrank*nlocal;
   if(iglobal<=n/2) 
    xtemp[ilocal] = iglobal*2.0/(n-1);  /* the rising part of the triangle*/   
    else 
    xtemp[ilocal] = (-2.0)*(iglobal-n/2)/(n-1)+1.;/* the declining part */
  }
    xtemp1[0]=xtemp[0];
/*            end of initialization               */

  for(i=0;i<nlocal;i++)
    xin[i] = xtemp[i];
  for(k=0;k<=maxiter;k++){
if(p !=1){
if (myrank!=p-1)
MPI.COMM_WORLD.Send(xtemp,nlocal-1,1,MPI.DOUBLE,myrank+1,0);
 if (myrank !=0)               
MPI.COMM_WORLD.Recv(xtemp1,0,1,MPI.DOUBLE,myrank-1,0);
if (myrank!=0)
MPI.COMM_WORLD.Send(xtemp,0,1,MPI.DOUBLE,myrank-1,0);
 if (myrank !=p-1)              
MPI.COMM_WORLD.Recv(xtemp,nlocal,1,MPI.DOUBLE,myrank+1,0);
           }
/* the solution scheme             */
/* solution is performed locally by each processor with 
   the required boundaries fed to it by the send and receive methods */
    for(ilocal=0;ilocal<nlocal;ilocal++){
      iglobal = ilocal+myrank*nlocal;
      if(iglobal!=0 && iglobal!=n-1){
        if(ilocal!=0)
          xout[ilocal] = tow*xtemp[ilocal+1]+2.*(1.-tow)*xtemp[ilocal]
            +tow*xtemp[ilocal-1]-xin[ilocal];
        else if (p!=1)
          xout[ilocal] = tow*xtemp[ilocal+1]+2.*(1.-tow)*xtemp[ilocal]
            +tow*xtemp1[0]-xin[ilocal];
      }
    }
/*  replacement of the new and old values, nescessary for the next time */ 
    for(i=0;i<nlocal;i++){
      xin[i]=xtemp[i];
      xtemp[i]=xout[i];
    }
          }


 
  for (i=0;i<nlocal;i++){
  iglobal = i + myrank * nlocal;
    System.out.print(iglobal);
    System.out.print("  ");
    System.out.print(xout[i]);
    System.out.println();
           }
MPI.Finalize();
             }
}

