/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <stdio.h>
#include <sys/fcntl.h>
#include <syslog.h>

#include "arg.h"

#define ROOT 0

int
usage(void) {
	syslog(LOG_WARNING, "jail [-u uid] [-g gid] [-l logfile] rootdir command [arg ...]\n");
	return 1;
}

int
main(int argc, char *argv[]) {
	int uflg=0, gflg=0;
	int uid, gid;
	char *logfn = 0;

	openlog("jail", LOG_PID, LOG_DAEMON);

	ARGBEGIN{
	case 'g':
		gid = atoi(ARGF());
		gflg++;
		break;
	case 'l':
		logfn = ARGF();
		break;
	case 'u':
		uid = atoi(ARGF());
		uflg++;
		break;
	default:
		return(usage());
	}ARGEND;

	if (argc < 2)
		return(usage());

	if (geteuid() != ROOT) {
		syslog(LOG_WARNING, "not super-user\n");
		fprintf(stderr, "jail: not super-user\n");
		return 2;
	}

	if (logfn) {
		int fd = open(logfn, O_WRONLY);
		if (fd < 0) {
			syslog(LOG_WARNING, "open log file: %m");
			perror("jail: open log file");
			return 3;
		}
		close(2);
		if (dup(fd) != 2) {
			syslog(LOG_WARNING, "dup failed: %m");
			return 4;
		}
		close(fd);
		lseek(2, SEEK_END, 0L);
	}

	if (chroot(argv[0]) < 0) {
		syslog(LOG_WARNING, "chroot failed: %m");
		perror("jail");
		return 5;
	}

	if(chdir("/") < 0) {
		syslog(LOG_WARNING, "chdir failed: %m");
		perror("jail: chdir");
		return 6;
	}

	if (gflg && setgid(gid) < 0) {
		syslog(LOG_WARNING, "setgid failed: %m");
		return 7;
	}

	if (uflg && setuid(uid) < 0) {
		syslog(LOG_WARNING, "setuid failed: %m");
		perror("jail: setuid");
		return 8;
	}

	execv(argv[1], &argv[1]);
	syslog(LOG_WARNING, "execv failed: %m");
	perror("jail: execv");
	return 9;
}
