/************************************************************************
 * Purpose:     A program to time send and return of a message. 
 *              Process 0 sends to all other processes, and each
 *              other process then sends back the same message.
 *
 * Author:      Steve Alexander
 *
 ************************************************************************/

#include	<stdio.h>
#include	<mpi.h>


#define ping 101
#define pong 101


void main ( int argc, char *argv[] )
{
  int ierror, rank, size, i, j, k, m, length, num_procs;
  int buffer[100000];
  double start, finish, time;
  MPI_Status status;

  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
  

   if (rank == 0)  printf("Length\tTotal Time\t number of processors\n");

   

    
     
      /*  Send buffer to all other processors using MPI_COMM_WORLD  
      *  Rank 0 receives a message of same length
      *  from all other processors */
      
       /* this loop lets us check the time for varying
         message lengths */
     for (length = 1; length <=  100000; length += 1000)
     {
     
      if (rank == 0) /* process 0 does this */
      {
       start = MPI_Wtime();/*get start time*/
       
       for(i = 1; i < num_procs; i++) /* send to, recieve from
                                         processors 1 through num_procs-1*/
       {
         
 
        for (j = 0; j < 100; j++)
        {
         MPI_Send(buffer, length, MPI_INT, i, ping, MPI_COMM_WORLD);
         

         MPI_Recv(buffer, length, MPI_INT, i, pong, MPI_COMM_WORLD, &status);
        }         
        
        
       }
       finish = MPI_Wtime(); /*get end time*/
   
       time = finish - start; /*get interval*/

       printf("%d\t%f\t%d\n", length, time/(100.0*num_procs), num_procs);
      }
     

     else /* all other processes do this */
     {
      for (m = 1; m < num_procs; m++)
      {
       if (rank == m)
       {
        for (k = 0; k < 100; k++)
        {       
         MPI_Recv(buffer, length, MPI_INT, 0, ping, MPI_COMM_WORLD, &status);

         MPI_Send(buffer, length, MPI_INT, 0, pong, MPI_COMM_WORLD);
        } 
       }
      }
     }            
    }
MPI_Finalize(); /*end MPI*/

}

