/*
*
*
*   author Steve Alexander version 1.0
*   
*   due 23 Feb 2000
*
*   Calculates the approximate value of the integral 
*   of the function dx / (sqrt(1 - (cos2 x)/2)) 
*   over the interval 0 to Pi/2. 
*
*/

#include <stdio.h>
#include "mpi.h"
#include <stdlib.h>
#include <math.h>

/* used to calculate local integral */
double Local(double local_leftx, double local_rightx, double deltax, int local_nintervals);

double f(double x); /* used to calculate the function of interest*/

 main(int argc, char** argv) 
 {

int p_rank;  /* process rank */
int p;       /* number of processes */
int nintervals=128;  /* number of trapezoids */
int local_nintervals; /* local number of trapezoids */
int source;  /* process sending integral */
int dest = 0; /* message destination is process 0 */
int tag = 0;  

double deltax; /* trapezoid base length */
double pi = 3.141592654;
double integral;  /* total integral */
double local_integral; /* local integral */
double leftx=0.0; /* left endpoint */
double local_leftx; /* local left endpoint */
double rightx=(pi/2.0); /* right endpoint */
double local_rightx; /* local right endpoint */

MPI_Status status;


MPI_Init(&argc, &argv);

MPI_Comm_rank(MPI_COMM_WORLD, &p_rank);  /* get this process rank */

MPI_Comm_size(MPI_COMM_WORLD, &p);  /* determine number of processors */

deltax = (0.5*pi)/(double) nintervals; /* trapezoid height is same for all processes */

local_nintervals = nintervals/p;  /* number of trapezoids same for all processes */

local_leftx = leftx + (((rightx - leftx)/(double) p) * (double) p_rank); /* gives local left endpoint */

local_rightx = local_leftx + ((rightx - leftx)/(double) p); /* gives local right endpoint */

/* calculates area of local integral section */
local_integral = Local(local_leftx, local_rightx, deltax, local_nintervals);

/* add the integrals from each process */
if (p_rank == 0) /* only the first processor enters this block */
{
  integral = local_integral;

  printf("Process 0 contributes: %7.5f to the total.\n",integral);

  for(source = 1; source < p; source++)
  {
    
    /* process 0 waits for other processes to send their info */
    MPI_Recv(&local_integral, 1, MPI_FLOAT, source, tag, MPI_COMM_WORLD, &status);
    
    printf("Process %d contributes: %7.5f to the total.\n",source,local_integral);
    
    integral = integral + local_integral; /* sums each process's part of the integral */
  }
}
else  /* all other processes send their integral info */
{
  MPI_Send(&local_integral, 1, MPI_FLOAT, dest, tag, MPI_COMM_WORLD);
}

/* print your results for the total integral */
if (p_rank == 0)
{
  printf("With n = %d trapezoids, the estimate\n", nintervals);
  printf("of the integral from %f to %f = %f\n", leftx, rightx, integral);
}

/* Shut down MPI */
MPI_Finalize();

}  /* end main */


/* function to estimate each section of the integral */
double Local(double local_leftx, double local_rightx, double deltax, int local_nintervals)
 {
   double loc_integral;
   double fleftx;
   double frightx; 
   int i;

   
   loc_integral = 0.0; /* value of the integral starts at zero */
   fleftx = f(local_leftx); /* gets value of f(left endpoint) for this section of the integral */
   for (i=1; i<=local_nintervals; i++) 
    {
     local_rightx = local_leftx + deltax * (double) i; /* gets local right endpoint */
     frightx = f(local_rightx); /* gets f(local right endpoint) */
     loc_integral += deltax * 0.5 *(fleftx + frightx); /* calculates area of trapezoid */
     fleftx = frightx; /* shift to the next trapezoid */
    }
   return loc_integral;  /* returns estimate of this process's part of the integral */
 }
 
/* function which calculates the function of interest on the argument supplied */
 double f(double x)
  {
   double y;
   
   y = cos(x);
   return (1.0/sqrt(1.0 - 0.5*y*y));
  }
