!Written in Fall96 as an assignment solution for CPS615.
!Author: Mike McMahon
!-------------------------------------------------------
!This program uses MPI to solve the heat equation over a rectangular
!grid.  The grid simulates a rectangular block of size 1 in. by 2 in.
!The boundary conditions are found using the heat equation
!supplied in an earlier homework, and the interior points set to -1.
!The alogrithm used is the Gauss - Seidel SOR using the
!red-black method.  The overall structure of the program  consists of
!            1)  Update red sites
! 			 2)  Communicate edge values (stored in EDGE array)
!			 3)  Update black sites using just calcutated reds
!			 4)  Communicate edge values again
!			 5)  Check tolerance (3.0e-6)
!			 6)  If tolerance is reached output # of iterations
!The distribution used below is similar to a block by rows in HPF
!This is easier to code than a checkerboard arrangement.  Don't
!worry, however, the code is almost 300 lines long. I did my
!fair share.  The number of iterations varies as omega (W)
!is changed in the range of 1<W<2.  The report lists the best
!values.
        Program sor
        Implicit none
        Include 'mpif.h'
        Real Ans(0:100,0:200)
        Real Oldans(0:100,0:200)
        real edge(0:100)
        Real tol,w,sum,iter
		Real tstart,tend
        Integer i,j,N,z
        Integer size,my_id,count,source,tag,comm,datatype
        Integer ierr,dest,k,partition
        Integer status(MPI_STATUS_SIZE)
        Integer hide(0:100,0:200)

        N=100
        Call MPI_INIT(ierr)
        Call MPI_COMM_RANK(MPI_COMM_WORLD,my_id,ierr)
        Call MPI_COMM_SIZE(MPI_COMM_WORLD,size,ierr)

        source=MPI_ANY_SOURCE
        tag=2001
        datatype=MPI_REAL
        Comm=MPI_COMM_WORLD
        count=(N+1)*(2*N+1)
        dest=size-1
        w=1.94
	write(6,*) 'When omega is:', ' ', w
!-------------set boundary conditions-------------

	If (my_id .eq. 0) then
        Do j=0,2*N
           Oldans(0,j)=-4.7 + 0.2*real(j/N) 
     $    - 0.3 + 0.09*((real(j/N))**2 -1)
     $      + log(sqrt((real(j/N)+0.2)**2 + 1.2**2))
        End do

        Do i=1,N
         Oldans(i,0)=-4.7-0.3*(1.0-real(i/N)) + 0.09*(-1.0+real(i/N))**2
     $      + log(sqrt(0.2**2 + (1.0-real(i/N)+0.2)**2))
        End Do

        Do i=1,N
           Oldans(i,2*N)=-4.7+0.2*2.0-0.3*(1.0-real(i/N))
     $    + 0.09*(4.0-(1.0-real(i/N))**2)
     $    + log(sqrt(2.2**2+(1.2-real(i/N))**2))
        End Do

        Do j=1,2*N-1
           Oldans(N,j)=-4.7+0.2*real(j/N)+0.09*(real(j/N))**2
     $      + log(sqrt((real(j/N)+0.2)**2 + 0.2**2))
        End Do

 
        Do i=1,N-1
           Do j=1,2*N-1
              Oldans(i,j)=-1.0
           End Do
         End Do

         do i=1,N-1
            do j=1,2*n-1
               hide(i,j) = 1
             end do
          end do

          do i=0,N
             do j=0,2*N
                ans(i,j)=oldans(i,j)
             end do
          end do
		 partition=2*n/size
	end if

        If (size .ne. 1) then
 
	do k=1,size-1
  		if(my_id .eq. 0) then
        Call MPI_SEND(Ans,count,datatype,k,100,comm,ierr)
        Call MPI_SEND(Oldans,count,datatype,k,100,comm,ierr)
         Call MPI_SEND(hide,count,datatype,k,100,comm,ierr)
		Call MPI_SEND(partition,1,MPI_INTEGER,k,100,comm,ierr)
	else if(my_id .eq. k) then 
	Call MPI_RECV(Ans,count,datatype,source,
     $        100,comm,status,ierr)
       	Call MPI_RECV(Oldans,count,datatype,source,
     $        100,comm,status,ierr)
       	Call MPI_RECV(hide,count,datatype,source,
     $        100,comm,status,ierr)
		Call MPI_RECV(partition,1,MPI_INTEGER,source,
     $       100,comm,status,ierr)
        end if
	end do
        end if
        
        tol=1.0
		iter=0.0
		tstart=mpi_wtime()
!***********start main loop**********************************
        Do z=1,800

!********************red phase!************************
        
                 do i=0,N
                    do j=my_id*partition+1,partition+my_id*partition
               if (mod((i+j),2) .eq. 0 .and. hide(i,j) .eq.1 ) then
                   ans(i,j)=w/4.0 * (Oldans(i+1,j) + Oldans(i-1,j) +
     $             Oldans(i,j+1)+Oldans(i,j-1)) + (1.0-w)*Oldans(i,j)
                       end if
                     end do
                  end do
!*****************communicate edge values************
		If (size .ne. 1 ) then
                  DO k=1,size-1
                    if (my_id .eq. 0 .and. k .eq. 1) then
                     do i=0,N
                        edge(i)=ans(i,my_id*partition+partition)
                      end do
                    call MPI_SEND(edge,101,datatype,k,
     $        100,comm,ierr)
                    end if

                   if (my_id .eq. k .and. k .ne. size-1) then
                     	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                       do i=0,N
                        edge(i)=ans(i,my_id*partition+partition)
                      end do
                       call MPI_SEND(edge,101,datatype,k+1,
     $        100,comm,ierr)                      
                       end if
                       
                 if (my_id .eq. size-1 .and. k .eq. size-1) then
                    	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                  end if
                  end do

              Do k=size-2,0,-1
                 if (my_id .eq. size -1 .and. k .eq. size-2) then
                    do i=0,N
                       edge(i) = ans(i,my_id*partition+1)
                    end do
                     call MPI_SEND(edge,101,datatype,k,
     $        100,comm,ierr)
                 end if

                 if (my_id .eq. k .and. my_id .ne. 0) then
                  	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                    do i=0,N
                       edge(i) = ans(i,my_id*partition+1)
                    end do
                    
                  call MPI_SEND(edge,101,datatype,k-1,
     $        100,comm,ierr)      

                end if

              if (my_id .eq. 0 .and. k .eq. 0) then
                	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
              end if

              end do
		end if
!*************************done with comm*****************

!***************into the black********************
                   do i=0,n
                    do j=my_id*partition+1,partition+my_id*partition
              if (mod((i+j),2) .ne. 0 .and. hide(i,j) .eq. 1) then
                      ans(i,j)=w/4.0 * (ans(i+1,j) + ans(i-1,j) +
     $                 ans(i,j+1)+ans(i,j-1)) + (1.0-w)*ans(i,j)
                        end if
                     end do
                  end do

!*******************Now Shift Black Sites***************** 
		If (size .ne. 1) then   
              DO k=1,size-1
                    if (my_id .eq. 0 .and. k .eq. 1) then
                     do i=0,N
                        edge(i)=ans(i,my_id*partition+partition)
                      end do
                    call MPI_SEND(edge,101,datatype,k,
     $        100,comm,ierr)
                    end if

                   if (my_id .eq. k .and. k .ne. size-1) then
                     	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                       do i=0,N
                        edge(i)=ans(i,my_id*partition+partition)
                      end do
                       call MPI_SEND(edge,101,datatype,k+1,
     $        100,comm,ierr)                      
                       end if
                       
                 if (my_id .eq. size-1 .and. k .eq. size-1) then
                    	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                  end if
                  end do

              Do k=size-2,0,-1
                 if (my_id .eq. size -1 .and. k .eq. size-2) then
                    do i=0,N
                       edge(i) = ans(i,my_id*partition+1)
                    end do
                     call MPI_SEND(edge,101,datatype,k,
     $        100,comm,ierr)
                 end if

                 if (my_id .eq. k .and. my_id .ne. 0) then
                  	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
                    do i=0,N
                       edge(i) = ans(i,my_id*partition+1)
                    end do
                    
                  call MPI_SEND(edge,101,datatype,k-1,
     $        100,comm,ierr)      

                end if

              if (my_id .eq. 0 .and. k .eq. 0) then
                	Call MPI_RECV(edge,101,datatype,source,
     $        100,comm,status,ierr)
              end if

              end do
		end if
!************** Check tolerance**************
                 do i=0,n
                 do j=my_id*partition+1,my_id*partition+partition
					if(j .ne. 2*N) THEN
                       sum=sum+abs(ans(i,j)-oldans(i,j))
					END IF
                     end do
                  end do  
		Do k=0,size-1
				if (my_id .eq. k .and. tol .ge. 3.0e-6) then
                  tol=abs(sum/((N+1)*(partition)))
                  sum=0.0
				if (tol .le. 3.0e-6) then
		tend=mpi_wtime()
		write(6,*) my_id,' ','is done in ',iter,'iterations'
		write(6,*) my_id,' ','took',tend-tstart,'seconds to complete'
				end if
			end if
		end do
!*********reset old values to new since not exiting********
                  do i=0,N
                     do j=my_id*partition+1,partition+my_id*partition
                       oldans(i,j)=ans(i,j)
                     end do 
                end do
             	do k=0,size-1 
                if (my_id .eq. k ) then
                   iter=iter+1
                 end if
				end do
                end do
!*****************end of main loop*************************

        call MPI_FINALIZE(ierr)


        End 
 
