/*
 * Investicate how the time taken varies with the size of the message.
 * Assume the message is an array of some type.
 * Do two or the three runs with the various message sizes, i.e.
 * by varying the size of the array
 *
 */

#include  <stdio.h>
#include  <mpi.h>      

#define proc_A 0   /* Process 0 */
#define proc_B 1   /* Process 1 */
#define ping 101   /* Tag for SEND */
#define pong 101   /* Tag for RECEIVE */

/* The size of Message which is float type */
float buffer[100000];

void processor_A (void); /* Process 0 subroutine */
void processor_B (void); /* Process 1 subroutine */


/* Point to Point Communication Test between Process 0 and Process 1.
 * Consider Synchronized SEND and Blocking RECEIVE.
 * Variable size of messages (1 - 100000)
 * Print out the length of message, time , transfer rate
 */

void main ( int argc, char *argv[] )
{
     int ierror, rank, size;  /* rank size */ 

/* MPI initialize */
     MPI_Init(&argc, &argv);

/* Getting the rank from Communicator */
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);

     if (rank == proc_A) processor_A();
     else if (rank == proc_B) processor_B();

     MPI_Finalize();

}

/* According to the length of the messages, 
 * this routine measures the time.
 * Process 0 
 */

void processor_A( void )
{

    int i, ierror;
    MPI_Status status;

    double start, finish, time;

    extern float buffer[100000];

    int length;

    printf("Length\tTotal Time\tTransfer Rate\n");

/* The Variable size of the messages */

     for (length = 1; length <=  100000; length += 1000){ 

        start = MPI_Wtime(); /* MPI time measurement */


        for (i = 1; i <= 100; i++){

            MPI_Ssend(buffer, length, MPI_FLOAT, proc_B, ping,
                      MPI_COMM_WORLD);

            MPI_Recv(buffer, length, MPI_FLOAT, proc_B, pong,
                     MPI_COMM_WORLD, &status);

        }

        finish = MPI_Wtime();

        time = finish - start;

        printf("%d\t%f\t%f\n", length, time/200.,
               (float)(2 * 8 * 100 * length)/time);

    }

}


/* Process 1 routine 
 * According to the length of the messages,
 * Process 1 can receive ans send the messages
 *
 */

void processor_B( void )
{
    int i, ierror;
    MPI_Status status;

    extern float buffer[100000];

    int length;

/* The variable size of the messages */

    for (length = 1; length <= 100000; length += 1000) { 

            for (i = 1; i <= 100; i++) {

                MPI_Recv(buffer, length, MPI_FLOAT, proc_A, ping,
                         MPI_COMM_WORLD, &status);

                MPI_Ssend(buffer, length, MPI_FLOAT, proc_A, pong,
                         MPI_COMM_WORLD);

            }

    }

}


