/* Part3 : About Ring Topology 
 * Modify the above program in order to measure the time taken by a message 
 * to travel between two adjacent processes along the ring. 
 * What happens to your timings when you vary the number of processes 
 * in the ring? Do the new timings agree with those you made 
 * with the above ping-pong program? 
 *
 */


/* The Definition of MPI */
#include "mpi.h"

void main( int argc, char **argv )
{
	int rank_value;     /* send and return value to/from the process */
	int predecessor, successor, tag = 0;
	int my_rank, num_processes;
	double start, finish, time;

	MPI_Request send_request;  /* Request for SEND */
	MPI_Request recv_request;  /* Request for RECEIVE */

	MPI_Status status;

	/* start up MPI	*/
	MPI_Init( &argc, &argv );

	/* get process rank */
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes */
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	successor = ( my_rank + 1 ) % num_processes;
	predecessor = ( my_rank - 1 + num_processes ) % num_processes;
	rank_value = my_rank;

	/* Each process sends its rank to the next processor until they get their own rank back*/
	
	        /* set up the starting time */
	        start = MPI_Wtime();

		MPI_Isend( &rank_value, 1, MPI_INT, successor, tag, MPI_COMM_WORLD, &send_request );
		MPI_Irecv( &rank_value, 1, MPI_INT, predecessor, tag, MPI_COMM_WORLD, &recv_request );

		MPI_Wait(&send_request, &status);
		MPI_Wait(&recv_request, &status);

		/* stop the time */
		finish = MPI_Wtime();
        
        time = finish - start;

        printf("Process\tTotal Time\n");

	printf( "%d\t%f\n", my_rank, time);

	/* close MPI*/
	MPI_Finalize();
}
