/*
 * Write a program in which the process with rank 0 broadcasts a message
 * via MPI_COMM_WORLD and then waits for every other process to send back
 * that message.
 *
 */

/* The definition of MPI */
#include "mpi.h"


void main( int argc, char **argv )
{
	int value = 41;	/* sending and return value to/from the process */
	int root = 0, tag = 0;
        int my_rank, num_processes;
	int i;

	double start, finish, time;
	MPI_Status status;

	/* start up MPI */
	MPI_Init( &argc, &argv );

	/* get process rank */
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes */
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	/* set up the starting time */
	start = MPI_Wtime();
  
	/* Process 0 broadcasts a message and then waits for every other process to send back that message. */
	MPI_Bcast( &value, 1, MPI_INT, root, MPI_COMM_WORLD );

	if( my_rank != root )
	{
	/* send back the message to the Process 0 */
		MPI_Send( &value, 1, MPI_INT, root, tag, MPI_COMM_WORLD );
	}
	else
	{
		/* receive the message Process 0 broadcast */
		for( i = 1; i < num_processes; i++ )
		{
			MPI_Recv( &value, 1, MPI_INT, i, tag, MPI_COMM_WORLD, &status );
		}
	}

	/* stop the time */
	finish = MPI_Wtime();

	time = finish - start;

	/* print out the result */
	if ( my_rank == root )
	{
	  printf( "Measurement Time with Broadcast = %f\n", time);
	}

	/* close MPI */
	MPI_Finalize();
}
