/* An Nbody type program 
 * This should have a similar form to the Nbody problem, but without the
 * complication of using something like Runge-Kutta to compute gravity between
 * each pair of bodies. Instead there will be a much simpler function to
 * compute for each pair of "bodies". 
 * In this case -- The Correlation Problem --

 ####################################################
 I did design an implementation of the "Much Better Message Parallel Algorithm"
 In this algorithm, I will first loop over the j's so that our outermost loop
 is to compute the correlations for the y's currently in the processor. Then
 it will communicate y's, so that another group of y's is brought to the
 processor. 
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/* The Definition of MPI */
#include "mpi.h"

#define PARTICLES 12000

/* Define the random number functions */
extern void srand48();
extern double drand48();

/* Main Program */
void main( int argc, char **argv )
{
	int predecessor, successor, tag = 0; /* source and destination processor variable */
	int my_rank, num_processes; /* rank and total processors */
	int root = 0;   /* process 0 */

	/* two sets of bodies of size PARTICLES */
	double x[PARTICLES], y[PARTICLES];

	/* correlations for computing */
	double correlation[PARTICLES];

	/* total value of each correlations */
	double total_correlation = 0.0;

	/* each correlations for each processor */
    double sub_correlation = 0.0;

	/* each processor should have PARTICLES/num_processes */
	int  local_p;
	int i,j,k;

	/* To make our correlation problem concrete, take this function */
	double f(double a, double b);
	
	MPI_Status status;

	/* start up MPI	*/
	MPI_Init( &argc, &argv );

	/* get process rank */
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes */
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

    /* each processor should have PARTICLES/num_processes */   
	local_p = PARTICLES/num_processes;

	/* Generate the initial values for each processor*/
	srand48(my_rank * 117); 
	for(i = 0; i < local_p; i++) {
		x[i] = drand48();
		y[i] = drand48();
		correlation[i] = 0.0;
	}

	/* right neighbors */
	predecessor = ( my_rank + 1 ) % num_processes;
	/* left neighbors */
	successor = ( my_rank - 1 + num_processes ) % num_processes;

	/* k ranges over processors */
	for(k = 0; k < num_processes;k++)
	{
		/* compute correlations of x's and y's in processor */
		for(j = 0; j < local_p; j++) /* j ranges over y's in processor */
		  for(i = 0; i < local_p; i++) /* i ranges over x's in processor */
			correlation[i] += f( x[i], y[j] );
		/* communicate y's */	
		MPI_Sendrecv_replace(y, local_p, MPI_DOUBLE, successor, tag, predecessor, tag, MPI_COMM_WORLD, &status);

	}

	/* sum correlation array in each processor */
	for(i = 0; i < local_p; i++)
		sub_correlation += correlation[i];

	/* each processor send result to processor 0, who gets total sum */
	if ( my_rank == root) {  
		total_correlation = sub_correlation;
			
        for(i = 1; i < num_processes; i++ )
        {
            MPI_Recv( &sub_correlation, 1, MPI_DOUBLE, i, tag, MPI_COMM_WORLD, &status );
			total_correlation += sub_correlation;
        }
	printf("Total Correlation is %f\n", total_correlation);

  	}
	else 
		MPI_Send(&sub_correlation, 1, MPI_DOUBLE, root, tag, MPI_COMM_WORLD );
		

	/* close MPI*/
	MPI_Finalize();
} /* MAIN */


/* To make our correlation problem concrete, take this function */
double f(double a, double b)
{
	return (abs(a*a + b*b));
}

