/* *************************************************************************
    ************************************************************************* */
   #include  <stdio.h>
   #include  <math.h>

   #include  "cortex.h"
   #include  "mpi.h"


/* =========================================================================
   |                      Structure of Layers                              |
   ========================================================================= */
   struct  input_layer {
           int    Digit_images [CATEGORY][TYPE][ROW][COLUMN];
           float  Digit_category [CATEGORY][CATEGORY];
           float  Digit [ROW][COLUMN]; /* input layer */
   } cell_i;

   struct  hidden_layer {
           float  activity [HUNIT];
           float  threshold [HUNIT];
           float  iweight [HUNIT][ITUPLE];    
           float  error_back [HUNIT];
           float  delta [HUNIT];
           float  delta_bias [HUNIT];
           float  delta_iweight [HUNIT][ITUPLE];          
   } cell_h;     /* hidden layer */

   struct  output_layer {
           float  activity [CATEGORY];
           float  desired_activity [CATEGORY];
           float  threshold [CATEGORY];
           float  weight [CATEGORY][HUNIT];
           float  error [CATEGORY];
           float  delta [CATEGORY];
           float  delta_bias [CATEGORY];
           float  delta_weight [CATEGORY][HUNIT];
   } cell_o;      /* output layer */

   struct  input_layer        *INPUT;            /* Input layer  */
   struct  hidden_layer       *HIDDEN;           /* Hidden layer */
   struct  output_layer       *OUTPUT;           /* Output layer */

/* =========================================================================
   |                    Functions and Variables                            |
   ========================================================================= */
   void    Read_Digit_image_category();
   void    Initialize_networks();
   void    Target_mapping();
   void    Learning_networks();
   void    Back_Propagation_rule();
   void    Storage_of_outputs();

   FILE    *stream1;                 /* Input file for Digit- images */
   FILE    *stream2;                 /* Input file for hunit's weights */
   FILE    *stream3;                 /* Input file for hunit's thresholds */
   FILE    *stream4;                 /* Input file for ounit's weights */
   FILE    *stream5;                 /* Input file for ounit's thresholds */
   FILE    *stream6;                 /* Output file for hunit's weights */
   FILE    *stream7;                 /* Output file for hunit's thresholds */
   FILE    *stream8;                 /* Output file for ounit's weights */
   FILE    *stream9;                 /* Output file for ounit's thresholds */
   FILE    *stream10;                /* Input file for Digit- category */

   int num_hidden_layer, num_output_layer; /* the number of hidden,output layer per each processor */
   int hidden_index_start, output_index_start;
   int hidden_index_end, output_index_end;
   int hidden_num_proc, output_num_proc;
 
   float temp_sum_square_error;
   float hidden_activity[HUNIT];
   float output_delta[CATEGORY];
   float output_weight[CATEGORY][HUNIT];
   float hidden_threshold[HUNIT];
   float output_threshold[CATEGORY];
   float hidden_iweight[HUNIT][ITUPLE];


main(int argc, char** argv)
{
    register  int   i;               /* Index */
    register  int   j;               /* Index */
    register  int   k;               /* Index */
    register  int   l;               /* Index */
    int       time = 0;              /* Processing step */

    int my_rank, num_processes;

    /* start up MPI	*/
    MPI_Init( &argc, &argv );

    /* get process rank */
    MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

    /* get number of processes */
    MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

    num_hidden_layer = (int) HUNIT / num_processes;
    num_output_layer = (int) CATEGORY / num_processes;

    output_num_proc = (int) HUNIT * num_output_layer;
    hidden_num_proc = (int) ITUPLE * num_hidden_layer;

    if(my_rank == 0) {
      printf("number of hidden layer = %d\n", num_hidden_layer);
      printf("number of output layer = %d\n", num_output_layer); 
      printf("proc of hidden layer = %d\n", hidden_num_proc);
      printf("proc of output layer = %d\n", output_num_proc);     
    }

    hidden_index_start = my_rank * num_hidden_layer;
    hidden_index_end = hidden_index_start + num_hidden_layer;

    output_index_start = my_rank * num_output_layer;
    output_index_end = output_index_start + num_output_layer;

    /* Creat and Open files */
      stream1 = fopen ( "project/digit.rf", "r" );
      stream2 = fopen ( "project/hidden.iwt", "r" );
      stream3 = fopen ( "project/hidden.ith", "r" );
      stream4 = fopen ( "project/output.iwt", "r" );
      stream5 = fopen ( "project/output.ith", "r" );
      stream10 = fopen ( "project/category.act", "r" );

      if(my_rank == 0) {
           stream6 = fopen ( "project/hidden.owt", "w" );
	   stream7 = fopen ( "project/hidden.oth", "w" );
	   stream8 = fopen ( "project/output.owt", "w" );
	   stream9 = fopen ( "project/output.oth", "w" );
      }

      INPUT = &cell_i;
      HIDDEN = &cell_h;
      OUTPUT = &cell_o;

    /* Read the Digit images and categories */
      Read_Digit_image_category();

    /* Initialize the SCHNN */
      Initialize_networks();

    /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
       +                                                                   +
       +                     Learning SCHNN                                +
       +                                                                   +
       +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
    /* STARTING */
      time = 0;

      do {
           /* What category ( 0 - 9 ) ? */
	for(i = 0; i < CATEGORY; i++){

           /* Mapping Digit's target patterns into output layer */
             Target_mapping(i);

           /* Random-orderly training of hand-written digits */
             for ( j = 0; j < TYPE; j++)
             {
                 /* Input token */
                   for ( k = 0; k < ROW; k++)
                   for ( l = 0; l < COLUMN; l++)
                   {
                     INPUT->Digit[k][l]= (float)INPUT->Digit_images[i][j][k][l];
                   }

                 /* Learning the SCHNN */
                   Learning_networks();
             }		     
	}

        /* Increment time */
	/*	if (my_rank == 0) printf("%d\n", time );  */
        time++;

      } while ( time <= 3000 ); /* Learning the SCHNN 3000*/

      MPI_Allgather( &(HIDDEN->threshold[hidden_index_start]), num_hidden_layer, MPI_FLOAT, hidden_threshold, num_hidden_layer, MPI_FLOAT, MPI_COMM_WORLD);

      MPI_Allgather( &(HIDDEN->iweight[hidden_index_start][0]), hidden_num_proc, MPI_FLOAT, &(hidden_iweight[0][0]), hidden_num_proc, MPI_FLOAT, MPI_COMM_WORLD);

      MPI_Allgather( &(OUTPUT->threshold[output_index_start]), num_output_layer, MPI_FLOAT, output_threshold, num_output_layer, MPI_FLOAT, MPI_COMM_WORLD);

      if (my_rank == 0) {
	/* Write the resulting outputs */
	Storage_of_outputs();
      }

     /* Close all files and Terminate the process */

      fclose(stream1);
      fclose(stream2);
      fclose(stream3);
      fclose(stream4);
      fclose(stream5);
      fclose(stream10);

      if (my_rank == 0) {
	fclose(stream6);
	fclose(stream7);
	fclose(stream8);
	fclose(stream9); }

      MPI_Finalize();

} /* main */

/* =========================================================================
   |                             Functions                                 |
   ========================================================================= */

/* Read the Digit images and categories */
void Read_Digit_image_category()
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;

    /* Read Digit-images */
      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < TYPE; j++)
      for ( k = 0; k < ROW; k++)
      for ( l = 0; l < COLUMN; l++)
      {
          fscanf ( stream1, "%1d", &INPUT->Digit_images [i][j][k][l] );
      }

    /* Read category patterns */
      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < CATEGORY; j++)
      {
          fscanf ( stream10, "%10f", &INPUT->Digit_category [i][j] );
      }

    return;
} /* Read_Digit_image_category */

/* Initialize the SCHNN */
void Initialize_networks()
{
    register  int   i;
    register  int   j;

    /* For hidden layer */
      for ( i = 0; i < HUNIT; i++)
      {
          fscanf ( stream3, "%10f", &HIDDEN->threshold [i] );
          for ( j = 0; j < ITUPLE; j++)
          {
              fscanf ( stream2, "%10f", &HIDDEN->iweight [i][j] );
          }
      }

    /* For output layer */
      for ( i = 0; i < CATEGORY; i++)
      {
          fscanf ( stream5, "%10f", &OUTPUT->threshold [i] );
          for ( j = 0; j < HUNIT; j++)
          {
              fscanf ( stream4, "%10f", &OUTPUT->weight [i][j] );
          }
      }

      return;
} /* Initialize_networks */

/* Mapping Digit's target pattern into output layer */
void  Target_mapping(i1)
register  int  i1;
{
    register  int   i;               /* Index */

    for ( i = 0; i < CATEGORY; i++ )
    {
        OUTPUT->desired_activity [i] = INPUT->Digit_category [i1][i];
    }

    return;
} /* Target_mapping */

/* Learning the SCHNN */
void Learning_networks()
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;
    float     wsum_hunit = 0.;
    float     wsum_ounit = 0.;
    float     netsum_hunit = 0.;
    float     netsum_ounit = 0.;

    /* Compute the output activities of hidden units */
      for ( i = hidden_index_start; i < hidden_index_end; i++)
      {
          for ( j = 0; j < ITUPLE; j++)
          {
              k = j / COLUMN;
              l = j % COLUMN;
              wsum_hunit += HIDDEN->iweight [i][j] * INPUT->Digit [k][l];
          }

          netsum_hunit = wsum_hunit + HIDDEN->threshold [i];
          HIDDEN->activity [i] = 1. / ( 1. + exp ( -1. * (double)netsum_hunit ) );
          wsum_hunit = 0.0;
      }

      MPI_Allgather( &(HIDDEN->activity[hidden_index_start]), num_hidden_layer, MPI_FLOAT, hidden_activity, num_hidden_layer, MPI_FLOAT, MPI_COMM_WORLD);

    /* Compute the output activities of output units */
      for ( j = output_index_start; j < output_index_end; j++)
      {
          for ( k = 0; k < HUNIT; k++)
          {
              wsum_ounit += OUTPUT->weight [j][k] * hidden_activity [k];
          }
          netsum_ounit = wsum_ounit + OUTPUT->threshold [j];
          OUTPUT->activity [j] = 1. / ( 1. + exp ( -1. * (double)netsum_ounit ) );
          wsum_ounit = 0.0;
      }

    /* Modifying all the weights and thresholds according to EBP rule */
      Back_Propagation_rule();

    return;
} /* Learning_networks */

/* Back_Propagation rule */
void Back_Propagation_rule()
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;
    float     momentum = 0.9;
    float     learning_rate = 0.1;

    /* For output units */
      for ( i = output_index_start; i < output_index_end; i++)
      {
          OUTPUT->error [i] = OUTPUT->desired_activity [i] - 
                              OUTPUT->activity [i];

          OUTPUT->delta [i] = OUTPUT->error [i] * OUTPUT->activity [i] * 
                              ( 1.0 - OUTPUT->activity [i]);

          OUTPUT->delta_bias [i] = learning_rate * OUTPUT->delta [i] + 
                                   momentum * OUTPUT->delta_bias [i];

          OUTPUT->threshold [i] += OUTPUT->delta_bias [i];
          for ( k = 0; k < HUNIT; k++)
          {
              OUTPUT->delta_weight [i][k] = learning_rate * 
                      OUTPUT->delta [i] * hidden_activity [k] + 
                      momentum * OUTPUT->delta_weight [i][k];
              OUTPUT->weight [i][k] += OUTPUT->delta_weight [i][k];
          }
      }

      MPI_Allgather( &(OUTPUT->delta[output_index_start]), num_output_layer, MPI_FLOAT, output_delta, num_output_layer, MPI_FLOAT, MPI_COMM_WORLD);

      MPI_Allgather( &(OUTPUT->weight[output_index_start][0]), output_num_proc, MPI_FLOAT, &(output_weight[0][0]), output_num_proc, MPI_FLOAT, MPI_COMM_WORLD);


    /* For hidden units */
      for ( i = hidden_index_start; i < hidden_index_end; i++)
      {
          HIDDEN->error_back [i] = 0.;
      }

      for ( i = hidden_index_start; i < hidden_index_end; i++)
      for ( j = 0; j < CATEGORY; j++)
      {
          HIDDEN->error_back [i] += output_delta [j] * 
                                    output_weight [j][i];
      }

      for ( i = hidden_index_start; i < hidden_index_end; i++)
      {
          HIDDEN->delta [i] = hidden_activity [i] * 
                              ( 1.0 - hidden_activity [i] ) * 
                              HIDDEN->error_back [i];
          HIDDEN->delta_bias [i] = learning_rate * HIDDEN->delta [i] + 
                                   momentum * HIDDEN->delta_bias [i];
          HIDDEN->threshold [i] += HIDDEN->delta_bias [i];
          for ( j = 0; j < ITUPLE; j++)
          {
              k = j / COLUMN;
              l = j % COLUMN;
              HIDDEN->delta_iweight [i][j] = learning_rate * 
                      HIDDEN->delta [i] * INPUT->Digit [k][l] + 
                      momentum * HIDDEN->delta_iweight [i][j];
              HIDDEN->iweight [i][j] += HIDDEN->delta_iweight [i][j];
          }
      }

    return;
} /* Back_Propagation_rule */


/* Write the resulting outputs, i.e. Weights and Thresholds */
void Storage_of_outputs()
{
    register  int   i;
    register  int   j;
    register  int   k;

    /* For hidden layer */
      for ( i = 0; i < HUNIT; i++)
      {
          fprintf ( stream7, "%10.6f", hidden_threshold [i] );
          if ( i == HUNIT - 1 )
          {
              fprintf ( stream7, "\n" );
          }
          for ( k = 0; k < ITUPLE; k++)
          {
              fprintf ( stream6, "%10.6f", hidden_iweight [i][k] );
	      if ( k == ITUPLE - 1 ) fprintf ( stream6, "\n" );
          }
      }

    /* For output layer */
      for ( i = 0; i < CATEGORY; i++)
      {
          fprintf ( stream9, "%10.6f", output_threshold [i] );
          if ( i == CATEGORY - 1 )  fprintf ( stream9, "\n" );
          for ( k = 0; k < HUNIT; k++)
          {
              fprintf ( stream8, "%10.6f", output_weight [i][k] );
              if ( k == HUNIT - 1 )  fprintf ( stream8, "\n" );
          }
      }

    return;
} /* Storage_of_outputs */
/* END */








