c  ASSIGN_2.f 
c
c****************************************************************************
c
c  CPS615 - ASSIGNMENT NO. 2
c
c  PURPOSE:  Numerical Integration Using MPI.
c
c****************************************************************************
c
       PROGRAM ASSIGN_2

       INCLUDE "mpif.h"

c Function to Integrate
	
c F(x) = {[1 - cos^2(x)/2]^(-0.5)} dx

c Boundaries for Integration = (0,pi/2)
	
c Variables

       DOUBLE PRECISION PI

       PARAMETER (PI = 3.141592653589793)

       DOUBLE PRECISION X, DX, LOWER, UPPER, SUM

       DOUBLE PRECISION INTBOUND(130), AVGY(130), AREA(130)

       INTEGER NINT, I, J

c X = interval boundary position
c DX = interval width
c LOWER = lower end of integration boundary
c UPPER = upper end of integration boundary
c SUM = for summing area
c INTBOUND() = array for beginning of trapezoids
c AVGY() = array for average trapezoid Y-value
c AREA() = array for trapezoid areas
c NINT = number of trapezoids
c I & J = loop indices

c Variables Needed for Parallel Processing

       INTEGER my_rank, p, source, dest, tag

       INTEGER status(MPI_STATUS_SIZE), ierr

       DOUBLE PRECISION PROC_NINT, PROC_LENGTH, PROC_LOWER, PROC_UPPER

       DOUBLE PRECISION GRANDTOT, start, finish, duration

c my_rank = process rank
c p = number of processes
c source = source processor; used as loop index
c dest = destination processor
c tag = tag for messages
c status = status of transmittal
c ierr = status of errors
c PROC_NINT = number of intervals for each processor
c PROC_LENGTH = length of integration boundary for each processor
c PROC_LOWER = lower bound for each processor
c PROC_UPPER = upper bound for each processor
c GRANDTOT = grand total area under curve

       CALL MPI_INIT(ierr)

       start = MPI_WTIME()

       CALL MPI_COMM_RANK(MPI_COMM_WORLD, my_rank, ierr)

       CALL MPI_COMM_SIZE(MPI_COMM_WORLD, p, ierr)

c Body of ASSIGN_2

c Establish known quantities.  
	
c Could change program, so these are entered.

       NINT = 128

       LOWER = 0

       UPPER = PI/2

       DX = (UPPER - LOWER) / NINT

       PROC_NINT = NINT / p

       PROC_LENGTH = (UPPER - LOWER) / p

       PROC_LOWER = LOWER + my_rank * PROC_LENGTH

       PROC_UPPER = PROC_LOWER + PROC_LENGTH

       dest = 0

       tag = 100

c On each processor, calculate interval areas and add them up.

       X = PROC_LOWER

       DO 10 I = 1, PROC_NINT + 1, 1
		
            INTBOUND(I) = 1/(SQRT(1 - 0.5*COS(X)**2))

            X = X + DX

10     CONTINUE

       SUM = 0

       DO 20 J = 1, PROC_NINT, 1

            AVGY(J) = 0.5*(INTBOUND(J) + INTBOUND(J+1))

            AREA(J) = AVGY(J)*DX

            SUM = SUM + AREA(J)

20     CONTINUE

       IF (my_rank .EQ. 0) THEN

            GRANDTOT = SUM

            Print *, "Processor 0 area =", SUM

            DO source = 1, p-1, 1

                 call MPI_RECV(SUM, 1, MPI_REAL, source, tag,
     Z           MPI_COMM_WORLD, status, ierr)

                 Print *, "Processor ", source, " Area =", SUM

                 GRANDTOT = GRANDTOT + SUM

            ENDDO

       ELSE 

            call MPI_SEND(SUM, 1, MPI_REAL, dest, tag,
     Z      MPI_COMM_WORLD, ierr)

       ENDIF

       IF (my_rank .EQ. 0) THEN

            PRINT *, "Grandtotal Area =", GRANDTOT

            finish = MPI_WTIME()

            duration = finish - start

            PRINT *, "Execution Time (sec) for ", p,
     Z      " Processors =", duration

       ENDIF

       CALL MPI_FINALIZE(ierr)

       end 