#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define NXMAX  50
#define NYMAX  50



#define MAXN 4

#define npx 2
#define npy 2

#define nptsx 16
#define nptsy 16

#define NDIM 2

main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, ibuf[MAXN], lsrc, ldest, rsrc, rdest;
    int dsrc, ddest, usrc, udest;
    int coords[NDIM], src;
    int ipx, ipy;


    double phi[NYMAX+1][NXMAX+1], oldphi[NYMAX+1][NXMAX+1];
    int mask[NYMAX+1][NXMAX+1];

    int value, size, i, j, iter, m, n, reorder;
    int dims[2], periods[2];
    int ntotx, ntoty, ntotx4, ntoty4;
    int nglobx, ngloby;


    MPI_Comm comm2d;
    MPI_Datatype  contig, strided;


    MPI_Init(&argc, &argv);

    MPI_Comm_rank( MPI_COMM_WORLD, &rank );

    MPI_Comm_size( MPI_COMM_WORLD, &size );

    ibuf[0] = npx;
    ibuf[1] = npy;
    ibuf[2] = nptsx;
    ibuf[3] = nptsy;

    MPI_Bcast (&ibuf, MAXN, MPI_INTEGER, 0, MPI_COMM_WORLD);

    /* .... set up communication ....*/

    periods[0] = 0;
    periods[1] = 0;

    dims[0] = npy;
    dims[1] = npx;

    reorder = 0;

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		    &comm2d);

    MPI_Cart_coords(comm2d, rank, 2, coords);

    ipy = coords[0];
    ipx = coords[1];

    printf("ipy = %d   ipx = %d \n", ipy, ipx);

    MPI_Cart_shift(comm2d, 0, 1, &dsrc, &ddest);
    MPI_Cart_shift(comm2d, 0, -1, &usrc, &udest);
    MPI_Cart_shift(comm2d, 1, 1, &rsrc, &rdest);
    MPI_Cart_shift(comm2d, 1, -1, &lsrc, &ldest);

    /* .... set up the grid ....*/

    for (j = 0; j < nptsx+1; j++){
      for (i = 0; i < nptsy+1; i++){
	phi[i][j] = 0.0;
	mask[i][j] = 1;
      }
    }
    if (ipx == 0) {
      for (i = 0; i < nptsy; i++)
	{
	  mask[i][1] = 0;
	}
    }

    if (ipx == (npx-1)){
      for (i = 0; i < nptsy; i++)
	{
	  mask[i][nptsx] = 0;
	}
    }

    if (ipy == 0) 
      {
	for (j = 0; j < nptsx; j++)
	  {
	    mask[1][j] = 0;
	  }
      }

    if (ipy == (npy-1)) 
      {
	for (j = 0; j < nptsx; j++)
	  {
	    mask[nptsy][j] = 0;
	  }
      }
    ntotx = npx * nptsx;
    ntoty = npy * nptsy;
  
    ntotx4 = ntotx / 4;
    ntoty4 = ntoty / 4;

    for (j = 0; j < nptsx; j++)
      {
	nglobx = (ipx * nptsx) + j;
	if ((nglobx > ntotx4) && (nglobx < (3 * ntotx4)))
	  {
	    for (i = 0; i < nptsy; i++)
	      {
		ngloby = ipy * nptsy + i;
		if ((ngloby > ntoty4) && (ngloby < (3 * ntoty4)))
		  {
		    phi[i][j] = 1.0;
		    mask[i][j] = 0;
		  }
	      }
	  }
      }

    MPI_Type_contiguous (nptsy, MPI_DOUBLE, &contig);
    
    MPI_Type_vector (nptsx, 1, NYMAX+2, MPI_DOUBLE, &strided);

    MPI_Type_commit (&contig);

    MPI_Type_commit (&strided);
    for (iter = 0; iter < 1000; iter++)
      {
      for (j = 0; j <= nptsx; j++)
	{
	  for (i = 0; i <= nptsy; i++)
	    {
	      oldphi[i][j] = phi[i][j];
	    }
	}
      MPI_Sendrecv (&(oldphi[1][1]), 1, contig, ldest, 30, 
		    &(oldphi[1][nptsx+1]), 1, contig, lsrc, 30,
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[1][nptsx]), 1, contig, rdest, 31,
		    &(oldphi[1][0]), 1, contig, ldest, 31, 
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[1][1]), 1, strided, udest, 40, 
		    &(oldphi[nptsy+1][1]), 1, strided, usrc, 40, 
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[nptsy][1]), 1, strided, ddest, 41,
		    &(oldphi[0][1]), 1, strided, dsrc, 41, 
		    comm2d, &status);

      for (j = 0; j < nptsx; j++)
	{
	  for (i = 0; i < nptsy; i++)
	    {
	      if (mask[i][j]) {
		phi[i][j] = 0.25 * (oldphi[i-1][j] + 
				    oldphi[i+1][j] 
				    + oldphi[i][j-1] + 
				    oldphi[i][j+1]);
	      }
	    }
	}
      }

    MPI_Type_free(&contig);
    MPI_Type_free(&strided);
    for (m = 0; m <= npx-1; m++){
      for (j = 0; j < nptsx; j++){
	for (n = 0; n <= npy-1; n++){
	  if ((m == ipx) && (n == ipy)) 
	    {
	      if (rank == 0)
		{
		  for (i = 0; i < nptsy; i++)
		    {
		      printf("%14.9e ", phi[i][j]);
		      if ((i % 5) == 0)
			printf("\n");
		    }
		  printf("\n");
		}
	      else
		{ 
		  MPI_Send(&(phi[1][j]), nptsy, MPI_DOUBLE, 
			   0, 50, MPI_COMM_WORLD);
		}
	    }
	  else
	    {
	      if (rank == 0) {
		coords[0] = n;
		coords[1] = m;
		MPI_Cart_rank (comm2d, coords, &src);
		MPI_Recv(&(oldphi[1][1]), nptsy, MPI_DOUBLE, 
			 src, 50, MPI_COMM_WORLD, &status);
		for (i = 0; i < nptsy; i++)
		  {
		    printf("%14.9e ", oldphi[i][1]);
		    if ((i % 5) == 0)
		      printf("\n");
		  }
		printf("\n");
	      }
	    }
	}
      }
    }
    MPI_Finalize();
}

