/* 
	. Process with rank 0 sends the same message to all other processes in MPI_COMM_WORLD 
	and then receives a message of the same length from all other processes. 
	. Investigate the time taken varies with the size of the messages.
*/

#include <stdio.h> 
#include <stdlib.h>
#include <math.h>
#include "mpi.h"

#define ping 101
#define pong 101

main(int argc, char** argv) {

    int each_process_rank, process_num,
		process_source, process_dest = 0,
	    length, i;
	float buffer[100000];
    double start, finish, time;
    MPI_Status  status;

	/* start up MPI
	*/
    MPI_Init(&argc, &argv);

	/* get process rank
	*/
    MPI_Comm_rank(MPI_COMM_WORLD, &each_process_rank);

	/* get number of processes
	*/
    MPI_Comm_size(MPI_COMM_WORLD, &process_num);

    if (each_process_rank == 0) 
	{
		printf("Length Elapsed_Time\n");

		for (length = 1; length <=  100000; length += 1000)
		{
			/* set up the starting time
			*/
        	start = MPI_Wtime();
        	for (i = 1; i <= 100; i++)
			{
        		for (process_source = 1; process_source < process_num; process_source++) 
				{
            		MPI_Send(buffer, length, MPI_FLOAT, process_source, ping, MPI_COMM_WORLD);
				}
				for (process_source = 1; process_source < process_num; process_source++) 
				{
            		MPI_Recv(buffer, length, MPI_FLOAT, process_source, pong, MPI_COMM_WORLD, &status);
        		}
		 	}

			/* stop the time
			*/
        	finish = MPI_Wtime();
        	time = finish - start;

			/* print out the result
			*/
        	printf("%d %f\n", length, time/200.);
    	}
    } 
	else 
	{  
   		for (length = 1; length <= 100000; length += 1000) 
		{
           	for (i = 1; i <= 100; i++) 
			{
              	MPI_Recv(buffer, length, MPI_FLOAT, process_dest, ping, MPI_COMM_WORLD, &status);
               	MPI_Send(buffer, length, MPI_FLOAT, process_dest, pong, MPI_COMM_WORLD);
           	}
   		}
    }

    MPI_Finalize();
} /*  main  */
