/*
	. Write a program in which the process with rank 0 broadcasts a message 
	via MPI_COMM_WORLD and then waits for every other process to send back 
	that message.
*/

#include "mpi.h"

void main( int argc, char **argv )
{
	int value = 10,			/* sending and return value to/from the process */
		root = 0, tag = 0,
		my_rank, num_processes;
	double begin_time, stop_time;
	MPI_Status status;

	/* start up MPI
	*/
	MPI_Init( &argc, &argv );

	/* set up the starting time
	*/
	begin_time = MPI_Wtime();
  
	/* get process rank
	*/
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes
	*/
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	/* Process 0 broadcasts a message and then waits for every other process 
	   to send back that message.
	*/
	MPI_Bcast( &value, 1, MPI_INT, root, MPI_COMM_WORLD );
	if( my_rank != root )
	{
		/* send back the message to the Process 0
		*/
		MPI_Send( &value, 1, MPI_INT, root, tag, MPI_COMM_WORLD );
	}
	else
	{
		/* receive the message Process 0 broadcast
		*/
		int i;
		for( i = 1; i < num_processes; i++ )
		{
			MPI_Recv( &value, 1, MPI_INT, i, tag, MPI_COMM_WORLD, &status );
		}
	}

	/* stop the time
	*/
	stop_time = MPI_Wtime();

	/* print out the result
	*/
	if ( my_rank == root )
	{
		printf( "When Bcast was used...\n" );
		printf( "Elapsed Time: %f\n", stop_time - begin_time );
	}

	/* close MPI
	*/
	MPI_Finalize();
}
