/*
	. Process with rank 0 scatters a message to all processes via MPI_COMM_WORLD 
	and then gathers back the message. 
*/

#include "mpi.h"

void main( int argc, char **argv )
{
	int value = 10, ret_val = -1,	/* sending and return value to/from the process */
		root = 0, tag = 0,
		my_rank, num_processes;
	double begin_time, stop_time;
	MPI_Status status;

	/* start up MPI
	*/
	MPI_Init( &argc, &argv );

	/* set up the starting time
	*/
	begin_time = MPI_Wtime();
  
	/* get process rank
	*/
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes
	*/
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	/* Process with rank 0 scatters a message to all processes 
	   and then gathers back the message. 
	*/
	MPI_Scatter( &value, 1, MPI_INT, &ret_val, 1, MPI_INT, root, MPI_COMM_WORLD );
	MPI_Gather( &value, 1, MPI_INT, &ret_val, 1, MPI_INT, root, MPI_COMM_WORLD );

	/* stop the time
	*/
	stop_time = MPI_Wtime();

	/* print out the result
	*/
	if ( my_rank == root )
	{
		printf( "When Scatter and Gather were used...\n" );
		printf( "Elapsed Time: %f\n", stop_time - begin_time );
	}

	/* close MPI
	*/
	MPI_Finalize();
}
