/************************************************************************
 * Purpose:     Each processor passes its rank to the processor on its 
 *              right. The processors keep passing the values they 
 *              receive until they receive their own rank. And then
 *              print out the sum of the values received.
 *
 * Author:      Samir Mehta
 * Date:        3/2/2000
 *
 ************************************************************************/

#include <stdio.h>
#include <mpi.h>

void main ( int argc, char *argv[] )
{
  int ierr, rank, size, i, lrank, rrank, lpass, rpass;
  int length, tag, sum = 0;
  MPI_Status status;

  ierr = MPI_Init(&argc, &argv);
  ierr = MPI_Comm_size(MPI_COMM_WORLD, &size);
  ierr = MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (rank == 0) { lrank = size - 1; }
  else { lrank = rank - 1; }

  if (rank == size - 1) { rrank = 0; }
  else { rrank = rank + 1; }

  tag = 100;

  ierr = MPI_Send(&rank, 1, MPI_INT, rrank, tag, MPI_COMM_WORLD);

  ierr = MPI_Recv(&lpass, 1, MPI_INT, lrank, tag, MPI_COMM_WORLD, &status);

  while (lpass != rank)
  {
      sum += lpass;
      rpass = lpass;

      ierr = MPI_Send(&rpass, 1, MPI_INT, rrank, tag, MPI_COMM_WORLD);

      ierr = MPI_Recv(&lpass, 1, MPI_INT, lrank, tag, MPI_COMM_WORLD, &status);

      if (lpass == rank) 
      {
	printf ("Rank %d received total values of %d\n",rank,sum);
	break;
      }
  }

  MPI_Finalize();

}
