/************************************************************************
 * Purpose:     Each processor passes its rank to the processor on its
 *              right. The processors keep passing the values they
 *              receive until they receive their own rank. The program
 *              print out the time taken by the message to travel
 *              between two adjacent processors.
 *
 * Author:      Samir Mehta
 * Date:        3/2/2000
 *
 ************************************************************************/

#include <stdio.h>
#include <mpi.h>

void main ( int argc, char *argv[] )
{
  int ierr, rank, size, i, lrank, rrank, lpass, rpass;
  int length, tag, sum = 0;
  MPI_Status status;
  double start, finish, lfinish, rfinish, time, ttltime = 0.0;

  ierr = MPI_Init(&argc, &argv);
  ierr = MPI_Comm_size(MPI_COMM_WORLD, &size);
  ierr = MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (rank == 0) { lrank = size - 1; }
  else { lrank = rank - 1; }

  if (rank == size - 1) { rrank = 0; }
  else { rrank = rank + 1; }

  tag = 100;

  start = MPI_Wtime();
  ierr = MPI_Send(&rank, 1, MPI_INT, rrank, tag, MPI_COMM_WORLD);

  ierr = MPI_Recv(&lpass, 1, MPI_INT, lrank, tag, MPI_COMM_WORLD, &status);
  lfinish = MPI_Wtime();

  ierr = MPI_Send(&lfinish, 1, MPI_DOUBLE, lrank, tag + 1, MPI_COMM_WORLD);

  ierr = MPI_Recv(&finish, 1, MPI_DOUBLE, rrank, tag + 1, MPI_COMM_WORLD, &status);

  ttltime += finish - start;

  while (lpass != rank)
  {
      sum += lpass;
      rpass = lpass;

      start = MPI_Wtime();
      ierr = MPI_Send(&rpass, 1, MPI_INT, rrank, tag, MPI_COMM_WORLD);

      ierr = MPI_Recv(&lpass, 1, MPI_INT, lrank, tag, MPI_COMM_WORLD, &status);
      lfinish = MPI_Wtime();

      ierr = MPI_Send(&lfinish, 1, MPI_DOUBLE, lrank, tag + 1, MPI_COMM_WORLD);

      ierr = MPI_Recv(&finish, 1, MPI_DOUBLE, rrank, tag + 1, MPI_COMM_WORLD, &status);

      ttltime += finish - start;

  }

  printf("Average time taken between P-%d and P-%d : %f\n", rank, rrank, ttltime/size);

  MPI_Finalize();

}
