#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define NXMAX  50
#define NYMAX  50



#define MAXN 4

#define xprocessors 2
#define yprocessors 2

#define xpoints 8
#define ypoints 8

#define NDIM 2

main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, input_buf[MAXN], left_source, left_dest, right_source, right_dest;
    int bottom_source, bottom_dest, top_source, top_dest;
    int coords[NDIM], src;
    int num_proc_x, num_proc_y;


    double phi[NYMAX+1][NXMAX+1], oldphi[NYMAX+1][NXMAX+1];
    int mask[NYMAX+1][NXMAX+1];

    int value, size, i, j, iter, m, n, reorder;
    int dims[2], periods[2];
    int ntotx, ntoty, ntotx4, ntoty4;
    int nglobx, ngloby;


    MPI_Comm comm2d;
    MPI_Datatype  contig, strided;


    MPI_Init(&argc, &argv);

    MPI_Comm_rank( MPI_COMM_WORLD, &rank );

    MPI_Comm_size( MPI_COMM_WORLD, &size );

    input_buf[0] = xprocessors;
    input_buf[1] = yprocessors;
    input_buf[2] = xpoints;
    input_buf[3] = ypoints;

    MPI_Bcast (&input_buf, MAXN, MPI_INT, 0, MPI_COMM_WORLD);

    /* .... set up communication ....*/

    periods[0] = 0;
    periods[1] = 0;

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    reorder = 0;

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		    &comm2d);

    MPI_Cart_coords(comm2d, rank, 2, coords);

    num_proc_y = coords[0];
    num_proc_x = coords[1];

    MPI_Cart_shift(comm2d, 0, 1, &bottom_source, &bottom_dest);
    MPI_Cart_shift(comm2d, 0, -1, &top_source, &top_dest);
    MPI_Cart_shift(comm2d, 1, 1, &right_source, &right_dest);
    MPI_Cart_shift(comm2d, 1, -1, &left_source, &left_dest);

    /* .... set up the grid ....*/

    for (j = 0; j < xpoints+1; j++){
      for (i = 0; i < ypoints+1; i++){
	phi[i][j] = 0.0;
	mask[i][j] = 1;
      }
    }
    if (num_proc_x == 0) {
      for (i = 0; i < ypoints; i++)
	{
	  mask[i][1] = 0;
	}
    }

    if (num_proc_x == (xprocessors-1)){
      for (i = 0; i < ypoints; i++)
	{
	  mask[i][xpoints] = 0;
	}
    }

    if (num_proc_y == 0) 
      {
	for (j = 0; j < xpoints; j++)
	  {
	    mask[1][j] = 0;
	  }
      }

    if (num_proc_y == (yprocessors-1)) 
      {
	for (j = 0; j < xpoints; j++)
	  {
	    mask[ypoints][j] = 0;
	  }
      }
    ntotx = xprocessors * xpoints;
    ntoty = yprocessors * ypoints;
  
    ntotx4 = ntotx / 4;
    ntoty4 = ntoty / 4;

    for (j = 0; j < xpoints; j++)
      {
	nglobx = (num_proc_x * xpoints) + j;
	if ((nglobx > ntotx4) && (nglobx < (3 * ntotx4)))
	  {
	    for (i = 0; i < ypoints; i++)
	      {
		ngloby = num_proc_y * ypoints + i;
		if ((ngloby > ntoty4) && (ngloby < (3 * ntoty4)))
		  {
		    phi[i][j] = 1.0;
		    mask[i][j] = 0;
		  }
	      }
	  }
      }

    MPI_Type_contiguous (ypoints, MPI_DOUBLE, &contig);
    
    MPI_Type_vector (xpoints, 1, NYMAX+2, MPI_DOUBLE, &strided);

    MPI_Type_commit (&contig);

    MPI_Type_commit (&strided);
    for (iter = 0; iter < 1000; iter++)
      {
      for (j = 0; j <= xpoints; j++)
	{
	  for (i = 0; i <= ypoints; i++)
	    {
	      oldphi[i][j] = phi[i][j];
	    }
	}
      MPI_Sendrecv (&(oldphi[1][1]), 1, contig, left_dest, 30, 
		    &(oldphi[1][xpoints+1]), 1, contig, left_source, 30,
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[1][xpoints]), 1, contig, right_dest, 31,
		    &(oldphi[1][0]), 1, contig, left_dest, 31, 
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[1][1]), 1, strided, top_dest, 40, 
		    &(oldphi[ypoints+1][1]), 1, strided, top_source, 40, 
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[ypoints][1]), 1, strided, bottom_dest, 41,
		    &(oldphi[0][1]), 1, strided, bottom_source, 41, 
		    comm2d, &status);

      for (j = 0; j < xpoints; j++)
	{
	  for (i = 0; i < ypoints; i++)
	    {
	      if (mask[i][j]) {
		phi[i][j] = 0.25 * (oldphi[i-1][j] + 
				    oldphi[i+1][j] 
				    + oldphi[i][j-1] + 
				    oldphi[i][j+1]);
	      }
	    }
	}
      }

    MPI_Type_free(&contig);
    MPI_Type_free(&strided);
    for (m = 0; m <= xprocessors-1; m++){
      for (j = 0; j < xpoints; j++){
	for (n = 0; n <= yprocessors-1; n++){
	  if ((m == num_proc_x) && (n == num_proc_y)) 
	    {
	      if (rank == 0)
		{
		  for (i = 0; i < ypoints; i++)
		    {
		      printf("%10.6e ", phi[i][j]);
		    }
		  printf("\n");
		}
	      else
		{ 
		  MPI_Send(&(phi[1][j]), ypoints, MPI_DOUBLE, 
			   0, 50, MPI_COMM_WORLD);
		}
	    }
	  else
	    {
	      if (rank == 0) {
		coords[0] = n;
		coords[1] = m;
		MPI_Cart_rank (comm2d, coords, &src);
		MPI_Recv(&(oldphi[1][1]), ypoints, MPI_DOUBLE, 
			 src, 50, MPI_COMM_WORLD, &status);
		for (i = 0; i < ypoints; i++)
		  {
		    printf("%10.6e ", oldphi[i][1]);
		  }
		printf("\n");
	      }
	    }
	}
      }
    }
    MPI_Finalize();
}

