#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>


#define MAXITER 1

#define xprocessors 2
#define yprocessors 2

#define xpoints 8
#define ypoints 8

#define NDIM 2

main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, leftneighbor, rightneighbor;
    int bottomneighbor, topneighbor;
    int coords[NDIM], src;
    int rankx, ranky;


    double phi[ypoints+2][xpoints+2], oldphi[ypoints+2][xpoints+2];
    int mask[ypoints+2][xpoints+2];

    int value, size, i, j, iter, m, n, reorder;
    int dims[2], periods[2];
    int ntotx, ntoty, ntotx4, ntoty4;
    int nglobx, ngloby;


    MPI_Comm comm2d;
    MPI_Datatype  contig, strided;


    MPI_Init(&argc, &argv);
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );


    /* .... set up communication ....*/

    periods[0] = 0;
    periods[1] = 0;

    reorder = 0;

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		    &comm2d);
    MPI_Cart_coords(comm2d, rank, 2, coords);

    ranky = coords[0];
    rankx = coords[1];

    MPI_Cart_shift(comm2d, 0, 1, &bottomneighbor, &topneighbor);
    MPI_Cart_shift(comm2d, 1, 1, &leftneighbor, &rightneighbor);

    /* .... set up the grid ....*/

    for (j = 0; j < xpoints+2; j++){
      for (i = 0; i < ypoints+2; i++){
	phi[i][j] = 0.0;
	mask[i][j] = 1;
      }
    }

    /* set mask to be false on boundary of original array */

    if (rankx == 0) {
      for (i = 1; i < ypoints+1; i++){ mask[i][1] = 0; phi[i][1]=1.0; }
    }

    if (rankx == (xprocessors-1)){  
      for (i = 1; i < ypoints+1; i++){ mask[i][xpoints] = 0;
                                     phi[i][xpoints]=1.0; }
    }

    if (ranky == 0)  {
	for (j = 1; j < xpoints+1; j++) { mask[1][j] = 0; phi[1][j]=1.0; }
      }

    if (ranky == (yprocessors-1)) 
      {	for (j = 1; j < xpoints+1; j++)  { mask[ypoints][j] = 0;
                                         phi[ypoints][j]=1.0;}
      }


    /*
    ntotx = xprocessors * xpoints;
    ntoty = yprocessors * ypoints;
  
    for (j = 0; j < xpoints; j++)
      {
	nglobx = (rankx * xpoints) + j;
	  {
	    for (i = 0; i < ypoints; i++)
	      {
		ngloby = ranky * ypoints + i;
		  {
		    solution[i][j] = function of nglobx and ngloby;
		    mask[i][j] = 0;
		  }
	      }
	  }
      }
   */

    MPI_Type_contiguous (ypoints, MPI_DOUBLE, &contig);
    MPI_Type_vector (xpoints, 1, ypoints+2, MPI_DOUBLE, &strided);
    MPI_Type_commit (&contig);
    MPI_Type_commit (&strided);


    for (iter = 0; iter < MAXITER; iter++)
      {
      for (j = 0; j <= xpoints+1; j++)
	{
	  for (i = 0; i <= ypoints+1; i++)
	    {
	      oldphi[i][j] = phi[i][j];
	    }
	}

      MPI_Sendrecv (&(oldphi[1][1]), 1, strided, leftneighbor, 30, 
		    &(oldphi[1][xpoints+1]), 1, strided, rightneighbor, 30,
		    comm2d, &status);

      MPI_Sendrecv (&(oldphi[1][xpoints]), 1, strided, rightneighbor, 31,
		    &(oldphi[1][0]), 1, strided, leftneighbor, 31, 
 		    comm2d, &status);

      printf ("rankx %d ranky %d oldphi[1][2]at b %lf \n", 
	      rankx, ranky, oldphi[1][2]);

#if 0
      MPI_Sendrecv (&(oldphi[1][1]), xpoints, MPI_DOUBLE, topneighbor, 40, 
		    &(oldphi[ypoints+1][1]), xpoints, MPI_DOUBLE, 
		    bottomneighbor, 40, comm2d, &status);
#endif

      MPI_Sendrecv (&(oldphi[1][1]), 1, contig, topneighbor, 40, 
		    &(oldphi[ypoints+1][1]), 1, contig, 
		    bottomneighbor, 40, comm2d, &status);


      printf ("rankx %d ranky %d oldphi[9][2]at c %lf \n", 
	      rankx, ranky, oldphi[9][2]);

      MPI_Sendrecv (&(oldphi[ypoints][1]), 1, contig, bottomneighbor, 41,
		    &(oldphi[0][1]), 1, contig, topneighbor, 41, 
		    comm2d, &status);


      for (j = 1; j <= xpoints; j++)
	{
	  for (i = 1; i <= ypoints; i++)
	    {
	      if (mask[i][j]) {
		phi[i][j] = 0.25 * (oldphi[i-1][j] + 
				    oldphi[i+1][j] 
				    + oldphi[i][j-1] + 
				    oldphi[i][j+1]);
	      }
	    }
	}
      }


printf ("rankx %d ranky %d phi[8][1]at av %lf \n", rankx, ranky, phi[8][1]);
printf ("rankx %d ranky %d mask[8][1]at av %d \n", rankx, ranky, mask[8][1]);
printf ("rankx %d ranky %d oldphi[8][1]at av %lf \n", rankx, ranky, oldphi[8][1]);
printf ("rankx %d ranky %d phi[8][2]at av %lf \n", rankx, ranky, phi[8][2]);
printf ("rankx %d ranky %d mask[8][2]at av %d \n", rankx, ranky, mask[8][2]);
printf ("rankx %d ranky %d oldphi[8][2]at av %lf \n", rankx, ranky, oldphi[8][2]);
printf ("rankx %d ranky %d phi[8][3]at av %lf \n", rankx, ranky, phi[8][3]);
printf ("rankx %d ranky %d mask[8][3]at av %d \n", rankx, ranky, mask[8][3]);
printf ("rankx %d ranky %d oldphi[8][3]at av %lf \n", rankx, ranky, oldphi[8][3]);
printf ("rankx %d ranky %d phi[9][2]at av %lf \n", rankx, ranky, phi[9][2]);
printf ("rankx %d ranky %d mask[9][2]at av %d \n", rankx, ranky, mask[9][2]);
printf ("rankx %d ranky %d oldphi[9][2]at av %lf \n", rankx, ranky, oldphi[9][2]);


    /* print results */

    for (m = 0; m <= xprocessors-1; m++){
      for (n = 0; n <= yprocessors-1; n++){
	MPI_Barrier(MPI_COMM_WORLD);

	if ((m == rankx) && (n == ranky)) 
	  for (j = 1; j <= xpoints; j++){
	    { printf("Proc rankx=%d ranky=%d, column=%d \n", m, n, j);
	    for (i = 1; i <= ypoints; i++)
	      { /* print jth column from current processor */
		printf("%10.6e ", phi[i][j]);
	      }
	    printf("\n");
	    fflush(stdout);
	    }
	  }
      }
    }

    MPI_Type_free(&contig);
    MPI_Type_free(&strided);

    MPI_Finalize();
}





