/*  Assignment 2 -- CPS 615
//  First MPI Assignment
//  Marjorie V Quant
//  quantm@rl.af.mil
//  13 Feb 00

//  Write an MPI program in which each processor is
//  assigned approximately the same number of contiguous
//  subintervals -- you may assume that the number of
//  sub-intervals (128) is exactly divisible by the number of
//  processors. For example, for the 4-process case, one
//  processor handles the first 32 subintervals, the second
//  processor the next 32 subintervals, and so on. In your
//  program each processor should evaluate its contribution
//  to the integral and output a line saying something like:
// (for example)Processor 3: contribution = 0.398278
//  Run your program on 2 and 4 nodes.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"


 main(int argc, char** argv)
 {

/*  Calling MPI_Init to set-up future MPI routines*/
   
   double Pi = 3.141592654;
   double a = 0.0;
   double b = Pi/2.0;	/*range for integral*/
   double base, area_x, TotalArea, a_x, b_x, x1, x2, y1, y2;
   double Totalarea_x;
   int N = 128;	/*number of total intervals*/
   int N_x;     /* local interval */
   int P;	/*number of processors used*/
   int rank;	/*rank (identity number)of the processor*/
   int i;
   int destination = 0, tag = 0, source;
   MPI_Status status;

   MPI_Init(&argc, &argv);
   
   /*  Identify the rank (processor number)*/
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);

   /*  How many processors are used */
   MPI_Comm_size (MPI_COMM_WORLD, &P);

   base = (b-a)/(double)N;	/*  base length for each trapazoid */
   N_x = N/P;		/*  number of intervals per processor */
   a_x = a+(double)rank*base*(double)N_x;	/*local left range value*/

   Totalarea_x = 0.0;
   x1 = a_x;
   x2 = a_x + base;
    
   /*Finds total area under curve for each processor*/
   for (i=0; i<N_x; i++)
   {
    y1 = (1/sqrt(1.0-0.5*cos(x1)*cos(x1)));
    y2 = (1/sqrt(1.0-0.5*cos(x2)*cos(x2)));

    area_x = ((base*y1)+(.5*base*(y2-y1)));
    Totalarea_x = Totalarea_x + area_x;
    x1 = x2;
    x2 = x2 + base;
   }

   printf("\nProcessor %d contributes %7.5f\n", rank, Totalarea_x);
  
   /*  Totals the area under the total curve*/
   if (rank == 0)
   {
     TotalArea = Totalarea_x;
     for (source = 1; source < P; source++)
     {
       MPI_Recv(&Totalarea_x, 1, MPI_FLOAT, source, tag,
     				MPI_COMM_WORLD, &status);
       TotalArea = TotalArea + Totalarea_x;
     }
   }
   else
   {
     MPI_Send(&Totalarea_x, 1, MPI_FLOAT, destination, tag,
              MPI_COMM_WORLD);
   }
    
    if(rank == 0) 
    {
     printf("Total area: %9.5f\n", TotalArea);
    }

  /*  Cleans up MPI */
   MPI_Finalize();

 }      /* ENDS Main*/