/* this program assumes the number of entities in each group can
   be divided by the number of processors */
/* By Qiang Zheng */

#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>

#define ROOT 0
#define NUM_OF_ITEMS 16

#define CORRELATION(a,b) ( ( ( a ) * ( a ) )+( ( b ) * ( b ) ) )

int main ( int argc, char *argv[] )
{
  /* The pointers to arrays of groups x, y, and intemediate result */
  double sum = 0.0, localSum = 0.0;
  double *localX, *localY, *result;
  /* the sub-arrays in each process */
  double *px, *py, *pResult; /* temporary pointers */
  int i=0, p, myId, myItemNum, q;
  MPI_Status status;
  /* p is the number of processors */

  /* setup the environment */
  MPI_Init( &argc, &argv );
  MPI_Comm_rank( MPI_COMM_WORLD, &myId );
  MPI_Comm_size( MPI_COMM_WORLD, &p );

/* debug information */
/* printf( "The number of nodes is : %d \n", p ); */

  myItemNum = NUM_OF_ITEMS/p;

/* printf( " Node %d has %d elements \n", myId, myItemNum );  */

  localX = ( double * ) malloc( sizeof( double ) * myItemNum );
  localY = ( double * ) malloc( sizeof( double ) * myItemNum );
  result = ( double * ) malloc( sizeof( double ) * myItemNum );
  /* allocate memory for localX, localY, intermediate result */

  if( localX == NULL || localY == NULL || result == NULL )
    return 1;

  /* Initialize the arrays */
  for ( i=0, px=localX; px != localX+myItemNum ; i++, px++ ) 
  {
     *px = (double) (i+(myId*myItemNum)) ; 
  }

  for ( i=0, py=localY ; py != localY+myItemNum ; i++, py++ ) 
  {
     *py = (double) (i+(myId*myItemNum) );
  }

  for ( i=0, pResult=result; pResult != result+myItemNum ; i++, pResult++ ) 
  {
     *pResult = 0.0 ;
  }

  /* q is the communication times, a process needs to communicate
     p times to get all y's values */
  for( q=p; q>0; q-- ) 
  {
    for( py=localY; py!=localY+myItemNum; py++ ) 
    {
      for ( px=localX, pResult=result; px!=localX+myItemNum ; px++, pResult++ ) 
      {
        *pResult += CORRELATION( *px,*py );
        /* 
        printf( " py = %f \n ", *py );
        printf( " px = %f \n ", *px );
        printf( " result = %f \n ", *pResult );
        */
      }
    }
    MPI_Sendrecv_replace( localY, myItemNum, MPI_DOUBLE, (myId+1)%p,
      0, (myId+p-1)%p, 0, MPI_COMM_WORLD, &status );
  }

  pResult = result+myItemNum;
  while ( --pResult >= result ) 
  {
    localSum += *pResult;
  }

  MPI_Reduce(&localSum, &sum, 1, MPI_DOUBLE
    , MPI_SUM, ROOT, MPI_COMM_WORLD );

  if ( myId == ROOT )
    printf( " The result of correlation is : %f \n", sum );

  free( localX );
  free( localY );
  free( result );
  MPI_Finalize();
  return 0;
}

