!!! Saleh Elmohamed,  saleh@npac.syr.edu     1996 - NPAC.
!!! ----------------------------------------------
!!! The below subroutine gauss_elimination solves the input linear equations
!!! via gaussian elimination with partial pivoting.
!!! Input: A set of coefficients of this linear system forming a
!!! matrix, and the number of equations of the system. This number is
!!! also the matrix size.
!!!
!!! The main program (gauss) ....

       program gauss

       interface
         subroutine gauss_elim(c,n)
           integer n
           real*8, dimension(:,:) :: c
!HPF$  distribute *(CYCLIC,*)   :: c
         end subroutine
       end interface

!!! Here, we will assume that 20 is the max number of equations that
!!! we will deal with ...
       
       parameter (n = 20)
       integer i
       real*8, dimension(n, n+1)::augmented
       double precision start_time, end_time
       real*8 X(n)
       character*50 filename


!HPF$ DISTRIBUTE augmented (CYCLIC, *)

!!! request name of input data file ...
!       filename = "inputdata.v2";
!       open (10, file=filename, status = "OLD")

       print *, ' Number of equations in the system: ', n
!!! read rows of the augmented matrix ...
!       do i = 1, n
!         read(10,*) augmented(i,1:n+1)
!       enddo
 
       call  Generate_A_x(augmented,X,n)
       print *, 'original X values: '
       print *, X

       print *, '==========================='
       print *, ' The original matrix size: ', n
       print *, '==========================='
       print *, ' Original augmented matrix is : '
       print *, '==========================='
!       print *, 'Input filename=', filename
       do i = 1, n
         print *, '<<<',i,'>>>'
         print *, augmented(i, 1:n+1)
       end do
       call timer(start_time)
       call gauss_elim(augmented, n)
       call timer(end_time)
       print *, ' Elapsed time = ', end_time-start_time
       print *, ' Solution Vector : ' 
!       print *, '==========================='
       do i=1, n
          print *, '<<<',i,'>>>'
          print *, augmented(i, 1:n+1)
       end do
       print *, '==========================='
       print *, 'Th final solution is ...'
       print *, augmented( 1:n, n+1)
       stop
       end
!**********************************************************
       subroutine timer(t)
          integer val(8)
          double precision t

          call date_and_time(values=val)
          t = dble(val(8))*1d-3 + dble(val(7)) +          &
          dble(val(6))*60d0 + dble(val(5))*3600d0

       end subroutine timer

!**********************************************************
!real function elapsed_time()
!      integer :: clock_count, clock_rate, clock_max
!      integer, save :: last_count
!      logical, save :: system_clock_not_yet_called = .true.
!
!     call system_clock(count=clock_count,&
!                       count_rate=clock_rate, &
!                       count_max=clock_max)
!      if (system_clock_not_yet_called) then
!        system_clock_not_yet_called = .false.
!         elapsed_time = 0.0
!      else if (clock_count > last_count) then
!         elapsed_time = real(clock_count-last_count)/real(clock_rate)
!     else
!         elapsed_time = real(clock_max-last_count+clock_count)&
!                 /real(clock_rate)
!      end if
!     last_count = clock_count
!      return
! end function elapsed_time
!**********************************************************
!**********************************************************
       subroutine gauss_elim(matrix,n)
         integer n
!         real*8, dimension (1:n, 1:n+1)  :: matrix
          real*8, dimension (:,:)  :: matrix
!HPF$ distribute *(cyclic,*)   :: matrix

! HPF$  INHERIT MATRIX

!!! local var ...
         integer i

!!! Also find_and_check_the_pivot, check_for_singularity, and back_sub
!!! are local subroutines to gauss_elim ...

!!! Do for as many equations in the linear system ... 

         do i = 1, n
           call find_and_check_the_pivot(matrix, i, n)
           call check_for_singularity(matrix, i, n)
         enddo
         
         call back_sub(matrix, n)

       end 

!**********************************************************
!**********************************************************
       subroutine find_and_check_the_pivot(a, i, nsize)
         integer i, nsize
         real*8, dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         real*8 , dimension (1:nsize+1) :: a_row
         integer pindex
         integer, dimension (1) :: pindex_locations

!!! the pivot locations is ...
         pindex_locations = maxloc(abs(a(i:nsize,i)))
         pindex = pindex_locations(1) + i - 1

         if(pindex /= i) then
            a_row = a(i,1:nsize+1)
            a(i,1:nsize+1) = a(pindex,1:nsize+1)
            a(pindex,1:nsize+1) = a_row
         endif
       end subroutine

!*******************************************************
!*******************************************************
! For more info on matrices and matrix singularity, see Numerical
! Recipes.

       subroutine check_for_singularity(a, i, nsize)
         integer i, nsize
         real*8 , dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         integer k

         if(abs(a(i,i)) < 1.0e-6) then
            write(*,*) 'error: matrix is singular ... bye bye.'
            stop
         endif
         a(i,i:nsize+1) = a(i,i:nsize+1)/a(i,i)
         do k = i+1, nsize
            a(k,i:nsize+1) = a(k,i:nsize+1) - a(k,i) * a(i,i:nsize+1)
         enddo
       end subroutine

!********************************************************
!********************************************************
! This routine will do simple back substitution ... (see Numerical
! Recipes book.) 
!
       subroutine back_sub(a, nsize)
         integer nsize
         real*8 , dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         integer k
         real*8  temp

         do k = nsize-1, 1,-1
            temp = sum(a(k,k+1:nsize) * a(k+1:nsize,nsize+1))
            a(k,nsize+1) = a(k,nsize+1) - temp
            a(k,k+1:nsize) = 0.0
         enddo
       end subroutine


!!! 
      subroutine Generate_A_x(A,X,n)
      real*8  A(n,n+1),X(n)
      integer n,i

      Call Random_number(A)
      Call Random_number(X)
      A = ((Aint(30*A)+1) / 10.0) - 1.0
      X = ((Aint(30*X)+1) / 10.0) - 1.0
      Do i=1,n
        A(i,n+1) = sum(A(i,1:n) * X) 
      enddo
      print *,'The just randomly generated matrix:' 
      print *, A
      return
      end subroutine
