!!! Saleh Elmohamed,  saleh@npac.syr.edu     1996 - NPAC.
!!! ----------------------------------------------
!!! The below subroutine gauss_elimination solves the input linear equations
!!! via gaussian elimination with partial pivoting.
!!! Input: A set of coefficients of this linear system forming a
!!! matrix, and the number of equations of the system. This number is
!!! also the matrix size.
!!!
!!! The main program (gauss) ....

       program gauss

       interface
         subroutine gauss_elim(c,n)
           integer n
           real*8, dimension(1:n, 1:n+1) :: c
         end subroutine
       end interface

!!! Here, we will assume that 20 is the max number of equations that
!!! we will deal with ...
       
       parameter (n = 20)
       integer i
       real*8, dimension(n, n+1)::augmented
!       real time1, elapsed_time
       character*50 filename


!HPF$ DISTRIBUTE augmented (BLOCK, *)

!!! request name of input data file ...
       filename = "pgi-files/inputdata.v2";
       open (10, file=filename, status = "OLD")

       print *, ' Number of equations in the system: ', n
!!! read rows of the augmented matrix ...
       do i = 1, n
         read(10,*) augmented(i,1:n+1)
       enddo
       print *, '==========================='
       print *, ' The original matrix size: ', n
       print *, '==========================='
       print *, ' Original augmented matrix is : '
       print *, '==========================='
       do i = 1, n
         print *, augmented(i, 1:n+1)
       end do
!       time1 =  elapsed_time()
       call gauss_elim(augmented(1:n, 1:n+1), n)
       print *, ' Solution Vector : ' 
       print *, '==========================='
       print *, augmented( 1:n, 1:n+1)
!       print *, ' Elapsed time = ', elapsed_time()

       end
!**********************************************************
!**********************************************************
!real function elapsed_time()
!      integer :: clock_count, clock_rate, clock_max
!      integer, save :: last_count
!      logical, save :: system_clock_not_yet_called = .true.

!      call system_clock(count=clock_count,&
!                       count_rate=clock_rate, &
!                       count_max=clock_max)
!      if (system_clock_not_yet_called) then
!         system_clock_not_yet_called = .false.
!         elapsed_time = 0.0
!      else if (clock_count > last_count) then
!         elapsed_time = real(clock_count-last_count)/real(clock_rate)
!      else
!         elapsed_time = real(clock_max-last_count+clock_count)&
!                  /real(clock_rate)
!      end if
!      last_count = clock_count
!      return
! end function elapsed_time
!
!**********************************************************
!**********************************************************
       subroutine gauss_elim(matrix,n)
         integer n
!         real*8, dimension (1:n, 1:n+1)  :: matrix
          real*8, dimension (:,:)  :: matrix

!HPF$  INHERIT MATRIX

!!! local var ...
         integer i

!!! Also find_and_check_the_pivot, check_for_singularity, and back_sub
!!! are local subroutines to gauss_elim ...

!!! Do for as many equations in the linear system ... 

         do i = 1, n
           call find_and_check_the_pivot(matrix, i, n)
           call check_for_singularity(matrix, i, n)
         enddo
         
         call back_sub(matrix, n)

       end 

!**********************************************************
!**********************************************************
       subroutine find_and_check_the_pivot(a, i, nsize)
         integer i, nsize
         real*8, dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         real*8 , dimension (1:nsize+1) :: a_row
         integer pindex
         integer, dimension (1) :: pindex_locations

!!! the pivot locations is ...
         pindex_locations = maxloc(abs(a(i:nsize,i)))
         pindex = pindex_locations(1) + i - 1

!!! adjust the pivot row ...
         if(pindex /= i) then
            a_row = a(i,1:nsize+1)
            a(i,1:nsize+1) = a(pindex,1:nsize+1)
            a(pindex,1:nsize+1) = a_row
         endif
       end subroutine

!*******************************************************
!*******************************************************
! For more info on matrices and matrix singularity, see Numerical
! Recipes.

       subroutine check_for_singularity(a, i, nsize)
         integer i, nsize
         real*8 , dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         integer k

         if(abs(a(i,i)) < 1.0e-6) then
            write(*,*) 'error: matrix is singular ... bye bye.'
            stop
         endif
         a(i,i:nsize+1) = a(i,i:nsize+1)/a(i,i)
         do k = i+1, nsize
            a(k,i:nsize+1) = a(k,i:nsize+1) - a(k,i) * a(i,i:nsize+1)
         enddo
       end subroutine

!********************************************************
!********************************************************
! This routine will do simple back substitution ... (see Numerical
! Recipes book.) 
!
       subroutine back_sub(a, nsize)
         integer nsize
         real*8 , dimension (1:nsize, 1:nsize+1) :: a

!!! local variables ...
         integer k
         real*8  temp

         do k = nsize-1, 1,-1
            temp = sum(a(k,k+1:nsize) * a(k+1:nsize,nsize+1))
            a(k,nsize+1) = a(k,nsize+1) - temp
            a(k,k+1:nsize) = 0.0
         enddo
       end subroutine
