/*---------------------------------------------------*/
/* File        : apple_pi.c                          */
/* Description : Calculates value of PI              */
/* Platform    : Sun Ultra Cluster                   */
/*               Virtual Programming Lab             */
/* Application : CPS615 Assignment #2, Fall '98      */
/*                                                   */
/* Author      : Subhash Nair (subhash@cat.syr.edu)  */
/*---------------------------------------------------*/
#include <stdio.h>
#include <mpi.h>
#include <math.h>

int main(int argc, char *argv[])
{	
  int numIntervals;  /*Number of intervals we will be intergrating*/
  int myid;          /*Process ID of this process*/ 
  int numProcs;      /*Total number of parallel processes*/ 
  int i;             /*Counter*/

  double local_pi;   /*To store the sum of the areas of the*/
                     /*intervals calculated by this process*/
  double global_pi[8];/*Array to gather all the local_pi values*/ 
  double pi=0.0;     /*To store the final value of PI*/
  double h, sum, x;  /*To store intermediate calculation results*/

  double PI25DT = 3.14159265358979323842643; /* Actual Value of PI for comparison*/
    
  MPI_Init(&argc, &argv);                   /*Initialize MPI*/
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs); /*Obtain the number of parallel processes*/
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);     /*Obtain the ID of this process*/

  if (myid == 0)
  {
    numIntervals = 50;  /*Process 0 decides the number of intervals for integration*/
  }

  /*Let all the processes know of the number of intervals*/
  MPI_Bcast(&numIntervals, 1, MPI_INT, 0, MPI_COMM_WORLD);

  /* Integral Calculation begins*/                                                       
  h = 1.0 /(double) numIntervals;  
  sum = 0.0;
  
  /* Each process integrates over a different set of specific intervals */
  /* Here lies the parallelization */
  for (i=myid+1; i<=numIntervals; i+=numProcs)
  {
     x = h *((double)i - 0.5);
     sum += 4.0/(1.0 + x*x);
  }

  /*Each process calculates the sum of the areas of the intervals it integrated over*/
  local_pi = h * sum;

  /*All values of local_pi are gathered into the global_pi[] array of the process 0*/
  MPI_Gather(&local_pi, 1, MPI_DOUBLE, &global_pi, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  /*Process 0 adds up all the local_pi values stored in it's global_pi[] array*/
  if (myid == 0)
  {
    for(i=0; i<numProcs; i++)
       pi += global_pi[i];
  }

  /*Let all the processes know of the final value of PI*/
  MPI_Bcast(&pi, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);

  printf("Process %d :\n\t PI = %.16f\n\t Error = %.16f\n\n", myid, pi, fabs(pi - PI25DT));


  MPI_Finalize;
}
