/*---------------------------------------------------*/
/* File        : apple_pi_squared.c                  */
/* Description : Calculates value of PI squared      */
/* Platform    : Sun Ultra Cluster                   */
/*               Virtual Programming Lab             */
/* Application : CPS615 Assignment #3, Fall '98      */
/*                                                   */
/* Author      : Subhash Nair (subhash@cat.syr.edu)  */
/*---------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

double sqrt();

/*The function that is being integrated*/
double f(double x, double y)
{
	return(16.0/((1.0 + x*x) * (1.0 + y*y)));
}

void main(int argc, char* argv[])
{
	double	PI=3.14159265358979323842643;
	int	p;		/*Number of processes*/
	int	n=20;		/*Number of intervals per process*/
	int	my_grid_rank;	/*The rank of this process in the Cartesian Grid*/
	MPI_Comm grid_comm;	/*Communicator that will contain all processes in MPI_COMM_WORLD*/
	int	dim_sizes[2];	/*Array that specifies the order of each dimension - the number of rows and colums*/
	int	wrap_around[2];	/*Array that specifies whether each dimesion is circular*/
	int	reorder = 1;	/*Reorders the processes in the new communicator*/
	int	coordinates[2];	/*Array that gives the coordinates of this process in the grid  */

	int	x1,y1,x2,y2;	/*Limits of integration for this node*/
	int	q;		/*Square root of the number of processes - used to construct the square grid*/
	double h;               /*Width of an interval*/
        double x, y;            /*Helper variables*/
        double pisquared, final_pisquared;

	int i,j;/*Counters*/

	/*Initializes MPI environment using command line arguments*/
	MPI_Init(&argc, &argv);

	/*Get the number of processes being used*/
	MPI_Comm_size(MPI_COMM_WORLD, &p);

	/*The square root of the number of processes = num. of rows and colums*/
	q = (int)sqrt(p);

	/*Set the order of each dimension to the number of rows and colums respectively*/
	dim_sizes[0] = dim_sizes[1] = q;

	/*Allow wrap-around*/
	wrap_around[0] = wrap_around[1] = 1;
   
	/*Create a new communicator grid_comm by caching a Cartesian topology with MPI_COMM_WORLD*/
	MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around, reorder, &grid_comm);

	/*Get the rank of this process*/
	MPI_Comm_rank(grid_comm, &my_grid_rank);
 
	/*Gets the coordinates of this process in the communicator*/
	MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

	h = (double)(1.0/(q*n));/*Width of an interval*/

        /*Set limits of integration for this process*/
	x1 = coordinates[0] * n;
	y1 = coordinates[1] * n;
	x2 = x1 + n;
	y2 = y1 + n;

        /*Compute the integral*/
	for(i = x1; i < x2; i++)
	{ 
		for (j = y1; j < y2; j++)
		{ 
			x = i * h;
			y = j * h;
			pisquared += (h*h) * (f(x+h/2.0, y+h/2.0));
		}
	}

	/*The root process collects the pi squared values from*/
        /*different processors and sums them up*/
	MPI_Reduce(&pisquared, &final_pisquared, 1, MPI_DOUBLE, MPI_SUM, 0, grid_comm);

        /*The root process prints out the final value of pi squared*/
	if(my_grid_rank == 0)
	{
		printf("PI squared = %.16f\n", final_pisquared);
                printf("Actual PI squared = %.16f\n", PI*PI);
		printf("Error = %.16f\n", fabs(final_pisquared - PI*PI));
                printf("Pi (from the computed) = %.16f, pi(from exact) = %.16f\n",
		       sqrt(final_pisquared), PI);
		
	}         
	MPI_Finalize();
}
