/*

 This program calculates the value of (PI * PI)

*/

/* Include all header files required */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

/* Declaration of the function vol which is defined as end */

double vol(double x, double y);

/* Start of the program */

main(int argc, char* argv[])
{
 int p;
 int dimensions[2], wrap[2], reorder = 1, coordinates[2];
 int rank ,q,i,j;
 double fin[10]; /* this stores the values of the pi squares calculated by the individual processors.*/
 double x, y;
 double initialx;             /* initial coordinate of x   */
 double finalx;               /* final coordinate of x     */
 double initialy;             /* initial coordinate of y   */
 double finaly;               /* final coordinate of y     */
 double h;
 double pisquare;    /* the value of pi square that each processor calculates                */
 double final = 0.0; /* The final value of pi square after all the values are added together */
 

 MPI_Comm grid_comm;

 /* This function starts up MPI, it must be called before any other MPI functio is called */
 MPI_Init(&argc, &argv);

 /* Returns the no of processors in the underlying comm */
    MPI_Comm_size(MPI_COMM_WORLD, &p);

    q = sqrt(p);
    dimensions[0] = dimensions[1] = q;
    wrap[0] = wrap[1] = 1;
 
 /* creates a new communicator with cartesian coordinate structure on 
    the processes*/  
    MPI_Cart_create(MPI_COMM_WORLD, 2, dimensions, wrap, reorder, &grid_comm);

 /* Returns the rank of the calling process in the group underlying the
    communicator comm */
    MPI_Comm_rank(grid_comm, &rank);

 /* Returns the coordinates of the process with rank rank in the
    cartesian communicator comm */ 
    MPI_Cart_coords(grid_comm, rank, 2, coordinates);
    

    h = 1/(double)p;
    initialx = (double)coordinates[0]/(double)q; /* Calculates the initial coordinate of x*/
    initialy = (double)coordinates[1]/(double)q; /* Calculates the initial coordinate of y*/
    finalx = (double)coordinates[0]/(double)q + (1/(double)q);  /* Calculates the final coordinate of x*/
    finaly = (double)coordinates[1]/(double)q + (1/(double)q);  /* Calculates the final coordinate of y*/
    for (x=initialx; x<finalx; x+=h)
      for (y=initialy; y<finaly; y+=h)
         pisquare += (h*h) * (vol(x,y) + vol(x,y+h) + vol(x+h,y) + vol(x+h,y+h))/4;
  
   pisquare = 16 * pisquare;
   
   MPI_Gather(&pisquare, 1, MPI_DOUBLE, &fin, 1, MPI_DOUBLE, 0, grid_comm);

  /* If the rank  of the process is 0 then it adds all the values and gives the final pi square value*/
   if(rank == 0)
   { 
      for(i = 0; i < p; i++)
        final += fin[i];
      printf("The  Final value of pi squared is %f\n", final); 
   }

   /* this function shuts down MPI. No MPI function may be called after this function */
   MPI_Finalize();
 }

 double vol(double x, double y)
 { 
  return (double)1/((1.0 +x*x) *(1.0 + y*y));
 }
