#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

int function_call(int m1, int m2);

main(int argc, char* argv[])
{ 
        int n , sendto, recvfrom, my_rank, x[6], y[6] ;

/* n is used for evenly distributing the N values to each processor  */
/* sendto - rank of process that the current process is sending to  */
/* recvfrom - rank of process from which current process receives data  */
/* my_rank - rank of the current process  */     
/* x[6], y[6] - arrays containing the two sets of measurements  */ 

        int local_correlation[6] , temp[6];    

/* local _correlation[6] - correlation array for each processor initialized to zero  */
/* temp[6] - temporary array used to receive y's from other processes  */

        int sum, correlation, p , i , j , k ;    

/* sum - used to sum up the elements of the correlation array for each processor  */
/*correlation -  the sum of all the correlation functions calculated by each process  */
/* p - the total number of processes in the communicator  */

	MPI_Comm grid_comm;    /*  communicator that will contain all processes in MPI_COMM_WORLD  */  
	MPI_Status status;
                
    /*  Initializes MPI environment using command line arguments  */
        MPI_Init(&argc, &argv);

    /*  Finds out how many processes are being used  */
        MPI_Comm_size(MPI_COMM_WORLD, &p);

    /*  Gets the rank of the process  */
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

        n = 6/p; 
        
   /* assigning random variables to x & y and initializing arrays */
        for (i=0; i<n;i++) 
	{
		 y[i] = my_rank*n + i;
		 x[i] = my_rank*n + i;
                 local_correlation[i] = 0;
                 temp[i]=0;
	}
        sum = 0; correlation=0;   /*  initializing to zero  */
	for (k=0; k<p; k++) 
	{	
           for ( j=0; j<n; j++)
	      for ( i=0; i<n; i++)             
		 local_correlation[i] += function_call(x[i], y[j]);
                               
	   sendto = (my_rank+1);
	   if (sendto >= p)
	   sendto = 0;
	   recvfrom = my_rank-1;
	   if (recvfrom < 0)
	      recvfrom = p-1;
		
	   MPI_Sendrecv(&y, n, MPI_INT, sendto ,0, &temp, n, MPI_INT, recvfrom ,0, MPI_COMM_WORLD, &status); 
                  
	   for(i = 0; i < n; i++)	
	      y[i] = temp[i];              
	}

     
        /*   summing up the elements of the correlation array  by each process  */
        for(i = 0; i < n; i++)
            sum += local_correlation[i];
			
        /*    a global reduction operation carried out to sum up the correlation functions obtained from all the processors  */
	MPI_Reduce(&sum, &correlation, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

        /*  Process 0 displays the correlation function obtained  */
	if (my_rank == 0)
		printf("\nThe output of correlation function is %d\n\n", correlation); 

	MPI_Finalize();
}

/*  The function that is used to compute the correlation  */
int function_call(int a, int b)
{
        return (a*a + b*b);
}
