#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define ALL MPI_COMM_WORLD
#define ANY MPI_ANY_SOURCE
#define TAG 54
#define DISP 58

int debug = 0;
int displayNorm = 1;

struct Blk
{
  double norm;
  double ** blkdata;
  int imax;
  int jmax;
  int tb;
  int te;
 
  double gnorm;
  int nblk;
  int nblks;
  double * sendbuff;
  double * sendRowbuff;
  double * sendColbuff;
  double * rbuff;
  double * lbuff;
  double * tbuff;
  double * bbuff;
  int lnb;   /*left neighbor */
  int rnb;
  int tnb;
  int bnb;
};

typedef struct Blk Block;
  

Block* BlockCreate(int rows, int cols)
{
	Block * blk;
        double ** grid;
	int i,j;

	grid = (double **)malloc(rows * sizeof(double *));
	for (i=0; i< rows; i++)
		grid[i] = (double*)malloc(cols * sizeof(double));

	blk = (Block *)malloc(sizeof(Block));
	blk->blkdata = grid;
	blk->imax = rows;
	blk->jmax = cols;
	
	blk->sendRowbuff = (double*)malloc(cols * sizeof(double));
	blk->sendColbuff = (double*)malloc(rows * sizeof(double));
	blk->rbuff = (double*)malloc(rows * sizeof(double));
	blk->lbuff = (double*)malloc(rows * sizeof(double));
	blk->tbuff = (double*)malloc(cols * sizeof(double));
	blk->bbuff = (double*)malloc(cols * sizeof(double));

	return(blk);
}

void BlockInit(Block * blk, int tBeg, int tEnd, int size, int rank)
{
  int i,j;
  double inorm;
  int rowNo;
  int NoOfBlocksInRow;
  int BlockRow;

  NoOfBlocksInRow = sqrt(size);
  BlockRow = rank/NoOfBlocksInRow;
  
  blk->tb = tBeg;
  blk->te = tEnd;
  blk->nblk = rank;
  blk->nblks = size;
  blk->lnb = (rank - 1);
  if (blk->lnb < BlockRow * NoOfBlocksInRow)
    blk->lnb = -1;
  
  blk->rnb = (rank + 1);
  if (blk->rnb >= BlockRow * NoOfBlocksInRow + NoOfBlocksInRow)
    blk->rnb = -1;

  blk->tnb = (rank - NoOfBlocksInRow);
  if (blk->tnb < 0)
    blk->tnb = -1;

  blk->bnb = (rank + NoOfBlocksInRow);
  if (blk->bnb >= size)
    blk->bnb = -1;
  
  blk->gnorm = 0;

  for(i=0; i< blk->imax; i++)
    for (j=0; j<blk->jmax; j++)
      blk->blkdata[i][j]=0.0;
  inorm = 0.0;
  
  if ( BlockRow == 0 )
    {
      for (j=0; j<blk->jmax; j++)
	{
	  blk->blkdata[0][j] = 10.0;
	  inorm += 10.0;
	}
    }


  if ( BlockRow == NoOfBlocksInRow -1 )
    {
      for (j=0;j<blk->jmax; j++)
	{
	  blk->blkdata[blk->imax-1][j]=10.0;
	  inorm += 10.0;
	}
    }


  if( rank%NoOfBlocksInRow == 0 )
    {
      for (i=0; i<blk->imax; i++)
	{
	  blk->blkdata[i][0]=10.0;
	  inorm += 10.0;
	}
      inorm -= 10.0;
    }


  if ( (rank+1)%NoOfBlocksInRow == 0 )
    {
      for (i=0; i<blk->imax; i++)
	{
	  blk->blkdata[i][blk->jmax-1]=10.0;
	  inorm += 10.0;
	}
      inorm -= 10.0;
    }

  blk->norm = inorm/(blk->imax * blk->jmax) ;
  
  if (debug)
    printf("BlockNo:%d Left:%d Right :%d Top :%d Bottom :%d\n",blk->nblk,blk->lnb,blk->rnb,blk->tnb,blk->bnb);

}




void BlockStep(Block * blk)
{
    int iend = blk->imax;
    int jend = blk->jmax;
    double temp1, temp2;
    double anorm=0.0;
    int i,j;
    
    int iStart, iEnd;
    int jStart, jEnd;
    double lVal=-1.0, rVal=-1.0, tVal=-1.0, bVal=-1.0;

 
    if (blk->lnb == -1)
      jStart = 1;
    else
      jStart = 0;
    
    if (blk->rnb == -1)
      jEnd = jend -1;
    else 
      jEnd = jend;

    if (blk->tnb == -1)
      iStart = 1;
    else 
      iStart = 0;

    if (blk->bnb == -1)
      iEnd = iend-1;
    else 
      iEnd = iend;

    if (debug)
      printf("Calculation Block# %d - iStart : %d iEnd : %d jStart : %d jEnd : %d\n",blk->nblk,iStart,iEnd,jStart, jEnd);
 
    for (i=iStart; i<iEnd; i++)
      {
	for (j=jStart; j<jEnd; j++)
	  {
	    lVal = -1; rVal = -1; tVal = -1; bVal = -1;
	    temp1 = blk->blkdata[i][j];
	    if (i == iStart)
	      {
		if (i==0)
		    tVal = blk->tbuff[j];
		else
		    tVal = blk->blkdata[i-1][j];
	      }

	    if (i==iEnd-1)
	      {
		if(i==iend-1)
		  bVal = blk->bbuff[j];
		else
		  bVal = blk->blkdata[i+1][j];
	      }

	    if (j==jStart)
	      {
		if (j==0)
		  lVal = blk->lbuff[i];
		else
		  lVal = blk->blkdata[i][j-1];
	      }

	    if (j== jEnd-1)
	      {
		if (j==jend-1)
		  rVal = blk->rbuff[i];
		else
		  rVal = blk->blkdata[i][j+1];
	      }

	    if (lVal == -1)
	      lVal = blk->blkdata[i][j-1];

	    if (rVal == -1)
	      rVal = blk->blkdata[i][j+1];
	    
	    if (tVal == -1)
	      tVal = blk->blkdata[i-1][j];
	    
	    if (bVal == -1)
	      bVal = blk->blkdata[i+1][j];

	    temp2 = (lVal + rVal + bVal + tVal + temp1)/5.0;
	    
	    blk->blkdata[i][j]= temp2;
	    anorm += fabs(temp2 - temp1);
	  }
      }

    blk->norm = anorm/(blk->imax * blk->jmax);
    blk->tb ++;
}


void BlockDestroy(Block * blk)
{
  int i;
  for (i=0; i<blk->imax; i++)
    {
       free(blk->blkdata[i]);
    }
  free (blk->blkdata);

  free(blk->sendRowbuff);
  free(blk->sendColbuff);
  free(blk->rbuff);
  free(blk->lbuff);
  free(blk->tbuff);
  free(blk->bbuff);
  free(blk);
}


void SendToBottom(Block * blk)
{
   int j;
   for(j=0; j<blk->jmax; j++)
	blk->sendRowbuff[j] = blk->blkdata[blk->imax-1][j];
   MPI_Send(blk->sendRowbuff, blk->jmax, MPI_DOUBLE, blk->bnb, TAG, ALL);
}

void SendToTop(Block * blk)
{
   int j;
   for(j=0; j<blk->jmax; j++)
	blk->sendRowbuff[j] = blk->blkdata[0][j];
   MPI_Send(blk->sendRowbuff, blk->jmax, MPI_DOUBLE, blk->tnb, TAG, ALL);	
		

}

void SendToLeft(Block * blk)
{
  int i;
  for (i=0; i<blk->imax; i++)
    {
      blk->sendColbuff[i] = blk->blkdata[i][0];
    }
  MPI_Send(blk->sendColbuff, blk->imax, MPI_DOUBLE, blk->lnb, TAG, ALL); 
}

void SendToRight(Block * blk)
{
  int i;
  for (i=0; i<blk->imax; i++)
    {
      blk->sendColbuff[i] = blk->blkdata[i][blk->jmax-1];
    }
  MPI_Send(blk->sendColbuff, blk->imax, MPI_DOUBLE, blk->rnb, TAG, ALL); 
}


void BlockNorm(Block * blk)
{
  MPI_Reduce(&blk->norm, &blk->gnorm, 1, MPI_DOUBLE, MPI_SUM, 0, ALL);
}


void SendTopAndLeft(Block * blk)
{
  if (blk->lnb != -1)
	SendToLeft(blk);

  if (blk->tnb != -1)
	SendToTop(blk);		
}

void SendBottomAndRight(Block * blk)
{
  if (blk->rnb != -1)
	SendToRight(blk);
  if (blk->bnb != -1)
	SendToBottom(blk);		
}	


void copyAll(double* dest, double * src, int elmts)
{
  int i;
  for (i=0; i<elmts; i++)
    dest[i]=src[i];
}


void Recv(Block *blk)
{

  int max_size;
  int i;
  double * buffer;
  int buff_size;
  int nb,src; 
  MPI_Status Status;
    
  if (blk->imax > blk->jmax)
    max_size = blk->imax;
  else
    max_size = blk->jmax;

  buffer = (double *)malloc(max_size * sizeof(double));

  for (i=0; i<4; i++)
    {
      if (i==0)
	nb = blk->rnb;
      if (i==1)
	nb = blk->bnb;
      if (i==2)
	nb = blk->lnb;
      if (i==3)
	nb = blk->tnb;

      if (nb != -1)
	{
	  MPI_Recv(buffer, max_size, MPI_DOUBLE,ANY, TAG, ALL, &Status);
	  src = Status.MPI_SOURCE; 

	  if (debug)
	    printf("Recd at : %d - from src :%d \n", blk->nblk, src);

	  if (src == blk->lnb)
	    copyAll(blk->lbuff, buffer, blk->imax);
	  
	  if (src == blk->rnb)
	    copyAll(blk->rbuff, buffer, blk->imax);

	  if (src == blk->tnb)
	    copyAll(blk->tbuff, buffer, blk->jmax);
	  
	  if (src == blk->bnb)
	    copyAll(blk->bbuff, buffer, blk->jmax);

	}
  
    }
  
  free(buffer);

}


void ShowGrid(Block * blk, int max_buff_rows, int max_buff_cols, int GridRows, int GridCols)
{
  int i,j,k,l,m;
  double *buff;
  double **data;
  int elmts;
  MPI_Status Status;
  int NoOfProcsInRow;
  int GridRowPos, GridColPos;
  int src;
  

  NoOfProcsInRow = sqrt(blk->nblks);


data = (double**)malloc(GridRows * sizeof(double*));
  for (i=0; i<GridRows; i++)
    data[i] = (double*)malloc(GridCols* sizeof(double)); 


   for (i=0; i<blk->imax; i++)
    for(j=0;j<blk->jmax;j++)
      data[i][j]=blk->blkdata[i][j];
      
 
  elmts = max_buff_rows * max_buff_cols + 2;     /* extra 2 for No of rows and No of cols */
  buff = (double*)malloc( elmts * sizeof(double));
 

  for (i=1; i<blk->nblks; i++)
    {
      MPI_Recv(buff, elmts, MPI_DOUBLE, ANY, DISP, ALL, &Status);
      src = Status.MPI_SOURCE; 
      
      GridColPos = (src%NoOfProcsInRow)*(GridCols/NoOfProcsInRow);
      GridRowPos = (src/NoOfProcsInRow)*(GridRows/NoOfProcsInRow);
     l=2;
      for (j=GridRowPos; j<(GridRowPos+(int)buff[0]); j++)
	{
	   for(k=GridColPos; k<(GridColPos+(int)buff[1]); k++)
	    {
	      data[j][k]=buff[l];
	      l++;
     	    }
	}
    }
  
      for (i=0; i<GridRows; i++)
	{
	  for(j=0;j<GridCols; j++)
	    printf("%2.2f ",data[i][j]);
	  printf("\n");
	}

  free(buff);

}

void SendGrid(Block * blk)
{
  double * buff;
  int i,j,k;
  int size;

  buff = (double*)malloc((blk->imax*blk->jmax+2) * sizeof(double));
  
  buff[0]=blk->imax;
  buff[1]=blk->jmax;
  
  k=2;
  for(i=0; i<blk->imax; i++)
    {
      for(j=0; j<blk->jmax; j++)
	{
	  buff[k]=(double)blk->blkdata[i][j];
	  k++;
	}
    }
  
 
  size = blk->imax * blk->jmax + 2;
  /* printf("Sedn size : %d\n",size); */
  MPI_Send(buff, size, MPI_DOUBLE, 0, DISP, ALL);
  free(buff);

}



void main(int argc, char * argv[])
{
  int sz, rnk;
  int tBegin, tEnd;
  int rows, cols, erows, ecols;
  int NoOfProcsInRow;
  Block *bp;
  int i,j;
  double begin_time, stop_time;
  int MaxRowsInProc, MaxColsInProc;

  MPI_Init(&argc, &argv);

  MPI_Comm_size(ALL, &sz);

  MPI_Comm_rank(ALL, &rnk);
  begin_time = MPI_Wtime();

  tBegin = 0;
  tEnd = 10;
  rows = 10;
  cols = 10;

  
  
  NoOfProcsInRow = sqrt(sz);
  
  if ((rnk%NoOfProcsInRow) == NoOfProcsInRow-1)
     ecols = cols - NoOfProcsInRow*(cols/NoOfProcsInRow);
  else
     ecols = 0;

  if ((rnk/NoOfProcsInRow) == NoOfProcsInRow -1)
    erows = rows - NoOfProcsInRow*(rows/NoOfProcsInRow);
  else
    erows = 0;

  if (debug)
    printf("Rank : %d Rows :%d Cols : %d\n",rnk,rows/NoOfProcsInRow + erows,   cols/NoOfProcsInRow+ ecols);

  bp = BlockCreate(rows/NoOfProcsInRow + erows, cols/NoOfProcsInRow+ ecols);
  
  BlockInit(bp, tBegin, tEnd, sz, rnk);

  BlockNorm(bp);
  if ((displayNorm) && (bp->nblk == 0))
	 printf("Global Norm ... %2.2f\n", bp->gnorm);
   
  while(  (bp->tb <= bp->te)  )
  {
       SendTopAndLeft(bp);
       Recv(bp);
       BlockStep(bp);
       SendBottomAndRight(bp);
       BlockNorm(bp);
       if ((displayNorm) && (bp->nblk == 0))
	 printf("Global Norm ... %2.2f\n", bp->gnorm);
  }
  
  MaxRowsInProc = rows - NoOfProcsInRow*(rows/NoOfProcsInRow) + rows/NoOfProcsInRow;
  MaxColsInProc = cols - NoOfProcsInRow*(cols/NoOfProcsInRow)+ cols/NoOfProcsInRow;

  if (bp->nblk == 0)
  {
    ShowGrid(bp,MaxRowsInProc, MaxColsInProc, rows, cols);
    stop_time = (MPI_Wtime() - begin_time);
    printf("... it took %8.5f seconds\n", stop_time);  
  }
  else
    SendGrid(bp);
   
  BlockDestroy(bp);
  
  MPI_Finalize();
}


