#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

/*The function that is being integrated*/
double f(double, double);
double sqrt(double);
double Calculate(int, int, int, double);

void main(int argc, char* argv[])
{
  int p;              /*Number of processes*/
  int a=0, b=1;       /*Intervals defined by a unit circle*/
  int n=20;           /*Number of intervals per process*/
  int my_grid_rank;   /*The rank of process in the Cartesian Grid*/
  int dim_sizes[2];   /*Order of each dimension*/
  int wrap_around[2]; /*To specify whether each dimension is circular*/
  int reorder = 1;    /*To reorder processes in the new communicator*/
  int coordinates[2]; /*Coordinates of this process in the grid  */
  int x1, y1;         /*Initial x and y values*/
  MPI_Comm grid_comm; /*Communicator with cartesian coordinate system*/
  double h, pisquared, final_pisquared=0, dptr[4];

  /*Basic MPI functions*/
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &p);

  /*Set order of each dimension*/
  dim_sizes[0]= sqrt(p);
  dim_sizes[1]= sqrt(p);

  /*Allow wrap-around*/
  wrap_around[0]= 1;
  wrap_around[1]= 1;
   
  /*Grid Topology MPI functions*/

  MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around, reorder, &grid_comm);
  MPI_Comm_rank(grid_comm, &my_grid_rank);
  MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

  h = (double)(b-a)/(double)(sqrt(p)*n); 
  x1= coordinates[0]* n;
  y1= coordinates[1]* n;

  pisquared= Calculate(x1, y1, n, h);
  
  /*Root processor adds pisquared values from other processes*/
  MPI_Gather(&pisquared, 1, MPI_DOUBLE, dptr, 1, MPI_DOUBLE, 0, grid_comm);   
  if(my_grid_rank==0)
  { 
    int i;
    for(i=0; i<p; i++)
      final_pisquared+= dptr[i];

    printf("Pi= %.15f, PI^2= %.15f\n", sqrt(final_pisquared), final_pisquared);
  }
   
  MPI_Finalize();
}

double f(double x, double y)
{
  return(1.0/((1.0 + x*x) * (1.0 + y*y)));
}

double Calculate(int x1, int y1, int n, double h)
{
  int i, j;
  double x, y, pisquared= 0;

  for(i=x1; i<(x1+n); i++)
  { 
    for (j=y1; j<(y1+n); j++)
    { 
      x= i * h;
      y= j * h;
      pisquared+= (h*h)* (f(x,y)+ f(x,y+h)+ f(x+h,y)+ f(x+h,y+h))/4;
    }
  }  
  
  return (pisquared*16);
}
