#include <mpi.h>
#include <stdio.h>
#include <math.h>

#define N 100

/* The given function */
double f(double x, double y)
{
  return fabs(x*x + y*y);
}

void main(int argc, char argv[]) 
{
  int n, p, rank, i, j, k, leftProc, rightProc;
  double *x, *y, local_correl, total_correl, buf;
  MPI_Status status;

  /* Basic MPI functions */
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &p);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  n = N/p;  /* number of values treated by each processor */
  x = (double*)malloc(sizeof(double) * n);
  y = (double*)malloc(sizeof(double) * n);

  /* initialize the two sequences */
  for(i=0; i<n; i++)
  {
    x[i]= 1;
    y[i]= 2;
  }

  local_correl = 0;

  leftProc  = (rank - 1 + p) % p;
  rightProc = (rank + 1) % p;

  /*compute local correlation*/
  for (k=0; k<p; k++) 
  {
    for (j=0; j<n; j++)
      for (i=0; i<n; i++)
        local_correl+= f(x[i], y[j]);

    MPI_Send(y, n, MPI_DOUBLE, leftProc , 1, MPI_COMM_WORLD);
    MPI_Recv(y, n, MPI_DOUBLE, rightProc, 1, MPI_COMM_WORLD, &status);
  }

  /*compute total correlation*/
  MPI_Reduce(&local_correl, &total_correl, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

  /* Processor 0 displays the total correlation */
  if(rank==0)
    printf("The correlation is: %.2lf\n", total_correl);

  MPI_Finalize();
}
