/***************************************************************
* Program : Calculation of the correlation function for an array 
*           of values.
* Project #4 -- CPS 615
*    
* Desc    : Program uses a circular buffer of MPI processes in a 
*           one-dimensional topology.
*
* Author  : Sachin R. Shanbhag
*****************************************************************/

#define TOTVALS 10
#include <stdio.h>
#include <mpi.h>
#include <math.h>

double corfunc( double a, double b ){
	return fabs( a * a + b * b );
}

int main(int argc, char *argv[])
{

	double	PI=3.14159265358979323842643;

	int numprocs, numvals, myid;
	
	int i, j, k, sendrank, recrank;
	
	double x[TOTVALS], y[TOTVALS], loccorrel, correlation;
	double locx[TOTVALS], locy[TOTVALS], temp[TOTVALS]; 
	
	MPI_Status status;

	/**************************************************************
	Initialize the MPI environment.
	***************************************************************/
	MPI_Init(&argc, &argv);

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	
	MPI_Comm_rank(MPI_COMM_WORLD, &myid);

	numvals = TOTVALS/numprocs;

	/**************************************************************
	Initialize the global storage array
	***************************************************************/
	for (i = 0; i <TOTVALS; i++){
		x[i] = i*i;
		y[i] = i*i;
	}

	/**************************************************************
	Assign each process a portion of the array.
	***************************************************************/
	for (i = 0; i < numvals; i++){
		locx[i] = x[myid * numvals +i];
		locy[i] = y[myid * numvals +i];
	}

	/**************************************************************
	Calculate the send and receive process ids for each process,
	***************************************************************/

	if ( myid == (numprocs-1) ) sendrank = 0;
	else sendrank = myid+1;

	if ( myid == 0 ) recrank = numprocs - 1;
	else recrank = myid-1;
	
	for (k=0; k < numprocs; k++){ /* loop the number of process times. */

	/**************************************************************
	Calculate the correlation for each process.
	***************************************************************/

		for ( j = 0; j < numvals; j++)
			for ( i = 0; i < numvals; i++)
				loccorrel += corfunc( locx[i], locy[j]);

	/**************************************************************
	Loop through the circular buffer of processes.
	***************************************************************/

		MPI_Sendrecv(&locy, (int)numvals, MPI_DOUBLE, sendrank ,0, &temp, (int)numvals, MPI_DOUBLE, recrank,0, MPI_COMM_WORLD, &status); 
		
		for(i = 0; i < numvals; i++)	
			locy[i] = temp[i];              
	}		

	/**************************************************************
	Reduce to a sum of all the values on the root process.
	***************************************************************/

	MPI_Reduce(&loccorrel, &correlation, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

	if (myid == 0)
		printf ( "Correlation value calculated = %f\n", correlation);
	

	MPI_Finalize;
}
