/***************************************************************
* Program : Calculation of sqauare PI by integration of 1/(1+x*x)
*           using MPI
* Project #3 -- CPS 615
*    
* Desc    : Program uses an approximation of the above integral
*           to calculate the value of square of PI. 
*
* Author  : Sachin R. Shanbhag
*****************************************************************/

#define INTERVAL 100
#include <stdio.h>
#include <mpi.h>
#include <math.h>

int main(int argc, char *argv[])
{

	double	PI=3.14159265358979323842643;

	int numprocs, myid, dims[2];
	int startx, starty, endx, endy, i, j;
	double h, x, y, finalpisq, midpisq;
	
	/**************************************************************
	Different variables to initialize the cartesian grid topology
	***************************************************************/

	MPI_Comm grid_comm;		/*Grid Communicator */
	int		dim_sizes[2];	/*order of each dimension */
	int		wrap_around[2];	/*wrapping flags for each dimension*/
	int		reorder = 1;	
	int		my_grid_rank;	/*rank of a process in the Cartesian Grid*/
	int		coordinates[2];	/*coordinates of a process in the grid  */

	
	/**************************************************************
	Initialize MPI, set up the grid communicator and get the rank
	of each process in the grid.
	***************************************************************/
	MPI_Init(&argc, &argv);

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);

	dim_sizes [0] = dim_sizes [1] = (int)sqrt(numprocs);
	wrap_around[0] = wrap_around[1] = 1;

	MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around, reorder, &grid_comm);

	MPI_Comm_rank(grid_comm, &my_grid_rank);

	if (my_grid_rank == 0){
		printf("***** Welcome to the Pi squared calculator *****\n");
		printf("Calculating .. Please wait .. \n");
	}

	MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

	
	/**************************************************************
	Give each process interval start and end values and calculate
	an interim value of the integral for each process (midpisq)
	***************************************************************/
	h = (double)(1.0/((int)sqrt(numprocs)*INTERVAL));

	startx = coordinates[0] * INTERVAL;
	starty = coordinates[1] * INTERVAL;
	endx = startx + INTERVAL;
	endy = starty + INTERVAL;

	midpisq = 0.0;
	for(i = startx; i < endx; i++)
		for (j = starty; j < endy; j++){ 
			x = i * h;
			y = j * h;
			midpisq += (h*h) * (16.0/((1.0 + x*x) * (1.0 + y*y)));
		}
	

	/**************************************************************
	Sum up all interim values and print out the results.
	***************************************************************/
	MPI_Reduce(&midpisq, &finalpisq, 1, MPI_DOUBLE, MPI_SUM, 0, grid_comm);

	
	if(my_grid_rank == 0){

		printf("\n\n");
		printf("**********************************************************\n");
		printf("Calculated value of Pi squared = %.16f\n", finalpisq);
		printf("Actual value of Pi squared = %.16f\n", PI*PI);
		printf("Error = %.16f\n", fabs(finalpisq - PI*PI));
		printf("**********************************************************\n");

	}         
 
	MPI_Finalize;

	return 0;
}
