/*========================================================================
   This example program starts up processes under MPI control.
   Process 0 acts as the collector for receiving messages from
   all other processes. Each message simply contains the process
   id of the sender.
==========================================================================*/

#include <stdio.h>
#include <mpi.h>

main( int argc, char **argv )
{
  int  i, my_rank, num_procs,
    buf_send[1], buf_recv[1], count,
    dest_proc, src_proc, msg_tag;

  MPI_Status       recv_status;

/*-----------------------*/
/* Executable Statements */
/*-----------------------*/

  MPI_Init( &argc, &argv );

  MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

  fprintf(stdout,"my_rank  = %d \n", my_rank );

  MPI_Comm_size( MPI_COMM_WORLD, &num_procs );

  fprintf(stdout,"I am proc %d of %d procs \n", my_rank, num_procs );
  fflush(stdout);

  count = 1;
  msg_tag = 111;
  if ( my_rank == 0 ) {
    for (i = 1; i < num_procs; i++ ) 
      {
	src_proc = i;
	MPI_Recv( &buf_recv[0], count, MPI_INT, src_proc, msg_tag,
		  MPI_COMM_WORLD, &recv_status );
	if ( buf_recv[0] != i ) 
	  {
	    fprintf(stdout,"receive error: bad data (%d)\n", buf_recv[0] );
	    MPI_Finalize();
	    exit(0);
	  }
	else
	  fprintf(stdout,"Received message from proc  %d \n", i );
      }
  }
  else 
    {
      buf_send[0] = my_rank;
      dest_proc = 0;
      MPI_Send( &buf_send[0], count, MPI_INT, dest_proc, msg_tag,
		MPI_COMM_WORLD );
      fprintf(stdout,"proc  %d  sent message \n", my_rank );
      fflush(stdout);
    }
  MPI_Barrier( MPI_COMM_WORLD );

  fprintf(stdout,"Shutdown MPI (proc  %d)\n", my_rank );

  MPI_Finalize();

  exit(0);
}
