#include "itemset.h"
#include "mpi.h"

/* allocate space for a new item set */
void newItemSet( ITEMSET * newSet, int setSize)
{
  newSet->size=setSize;
  newSet->idSet= ( int * ) malloc ( setSize*sizeof( int ) );
  newSet->currentIdx=0;
  newSet->support=0;
  newSet->recCount=0;
  newSet->recSet=NULL;
}

/* add an item into an item set */
void addAnItem( ITEMSET * set, int id )
{
  if( set->currentIdx>=set->size ) {
    printf("number of items greater than expected \n ");
    exit( 1 );
  }
  set->idSet[ ( set->currentIdx )++ ]=id;
}

/* judge if two item sets contain the same set of items */
int equalItemSet( ITEMSET * is0, ITEMSET * is1 )
{
  int i;

  /* if ( is0->size != is1->size ) return 0; */
  if ( is0->currentIdx != is1->currentIdx ) return 0;
  i=(is0->currentIdx);
  while ( ( i-->0 )&&( ( is0->idSet )[i]==( is1->idSet )[i] ) ) ;
  return ( ( i<0 ? 1 : 0 ) );
}

/* when two item sets match, it means:
 * 1) the item number they contain are same;
 * 2) each item in the set are identical except the last one
 * if two itemset match, then we combine them to a new itemset
 * which contain these two itemsets' joint set
 */
int combineItemSet( int * table, int colNum, ITEMSET * is0, ITEMSET * is1 , ITEMSET * is2 )
{
  int i, j, newRecNum, coOccurs, newId;
  int * newIdSet;
  int * newRecSet;

  if ( ( is0->size ) != ( is1->size ) )
    return 0;

  i=(is0->size);

  if( ( is0->idSet )[i-1]==( is1->idSet )[i-1] ) {
    return 0;
  }

  newIdSet = ( int * ) malloc ( ( ( is0->size)+1 )*sizeof( int ) ) ;
  is2->idSet=newIdSet;
  is2->size=(is0->size)+1 ;

  for ( i=0; i<( is0->size)-1; i++ ) {
    if ( ( is0->idSet )[i]==( is1->idSet )[i] ) {
      newIdSet[ i ]=( is0->idSet )[ i ];
    } else {
      break;
    }
  }

/* printf("i=%d\n",i); */

  if( i<(is0->size)-1 ) {
    free( newIdSet );
    is2->currentIdx=0;
    is2->idSet=NULL;
    return 0;
  }

  /* i==( ( is0->size )-1 ) here */

  newIdSet[ i ] = ( is0->idSet )[ i ];
  newIdSet[ i+1 ] = ( is1->idSet )[ i ];
  is2->currentIdx=i+1;
  is2->support++;
  is2->size=( is0->size )+1;

  /*
   * we don't know how many records contain the itemset in both
   * itemset 0 and itemset 1, but we know it will not exceed the
   * smaller number of records in either one, so we simplly request 
   * memory enough for the smaller set
   */
  newId=is2->idSet[ i ];
  newRecNum= ( is0->recCount<is1->recCount ? is0->recCount : is1->recCount );
/* printf("is0->recCount=%d\n",is0->recCount);
printf("is1->recCount=%d\n",is1->recCount);
for(i=0;i<is0->size;i++){
printf("is0->idset[%d]=%d\t",i, is0->idSet[i]);
printf("is1->idset[%d]=%d\n",i, is1->idSet[i]);
}

printf("newRecNum=%d\n",newRecNum);
*/
  newRecSet = ( int * ) malloc ( newRecNum*sizeof( int ) );

  /*
   * calculating how many times the item with id "newId"
   * occurs in the records of is0
   */
  coOccurs=occurTimes( table, colNum, is0, newId, newRecSet );
/* printf("coOccurs=%d\n",coOccurs); */
  if ( coOccurs==0 ) {
    free( newIdSet );
    is2->currentIdx=0;
    is2->idSet=NULL;
    free( newRecSet );
    is2->recCount=0;
    return 0;
  }

  is2->recCount=coOccurs;
  is2->recSet=newRecSet;
/* printf("coOccurs=%d\n",coOccurs);
for(i=0;i<is2->size;i++){
printf("is2->idSet[%d]=%d\n",i,is2->idSet[i]);
}
for( i=0;i<coOccurs; i++)
{
printf("is2->recSet[%d]=%d\n",i,is2->recSet[i]);
} */

  return 1;
}

int occurTimes( int * table, int colNum, ITEMSET* is0, int id, int * recSet )
{
  int coOccurTimes=0;
  int i;

  for( i=0; i<is0->recCount; i++ ) {
    if( table[ ( is0->recSet )[i]*colNum+id ] ) {
      coOccurTimes++;
      recSet[ i ]=( is0->recSet )[i];
    }
  }

  return coOccurTimes;
}

/*
 * count how many occurence happened in database
 * in the new item set
 * since is1 contains one more item than is0 
 * and item set of is0 is subset of is1,
 * we just need to count how many records
 * containing is0 also contains the last 
 * item in is1 
 */
int occurCount( int * table, int colNum, ITEMSET * is0, ITEMSET * is1 )
{
  int tempCount;
  int i, itemId;
  int * newRecSet;
  int j;

  itemId=is0->idSet[ ( is1->size )-1 ];
  i=is0->recCount;

/* printf(" is0's item number %d\n", is0->size );
 * printf("is0: (");

 * for( j= 0; j<is0->recCount; j++ ) {
 * printf("%d,",is0->recSet[j]);
 * }
 * printf(")");
 */

newRecSet = ( int * ) malloc( ( is0->recCount )*sizeof( int ) ) ;

printf(" count of is0 is %d \n", is0->recCount);

  is1->recSet = ( int * ) malloc( ( is0->recCount )*sizeof( int ) ) ;
  while( i-->0 ) {
    if ( table[ is0->recSet[i]*colNum+itemId ]>0 ) {
      ( is1->recCount )++;
      is1->recSet[i]=is0->recSet[i];
    }
  }

/* printf("tempCount=%d\n",tempCount); */
 return is1->recCount;
}


/* delete an item set */
void delItemSet( ITEMSET * set )
{
  free( set->idSet );
  set->idSet=NULL;
  set->size=0;
  set->currentIdx=0;
  set->support=0;
  set->recCount=0;
  free (set->recSet);
  set->recSet=NULL;
}

/*
 * allocate space for an item set list 
 * contain the same number of items
 */
void newItemSetList( ITEMSETLIST * itStLst, int itemNum, int size )
{
  itStLst->itemNum=itemNum;
  itStLst->listSize=size;
  itStLst->currentIdx=0;
  itStLst->setList=( ITEMSET * ) malloc ( size*sizeof( ITEMSET ) );
}

/* add an itemset in an item set list */
void addAnItemSet( ITEMSETLIST * list )
{
  if ( list->currentIdx>=list->listSize ) {
    printf("number of itemsets in a set list is greater than expected \n ");
    exit( 1 );
  }

  newItemSet( ( list->setList )+( list->currentIdx), list->itemNum );
  ( list->currentIdx )++;
}

/* remove an item set list */
void delItemSetList( ITEMSETLIST * list )
{
  ITEMSET * sp;
   
  sp=( list->setList )+( list->currentIdx );
  while( sp--!=list->setList ) {
    delItemSet( sp );
  }
  free( list->setList );
  list->setList=NULL;
}

/*
 * this list is used to contain the real min-itemset
 * list which means every item set in it is with the
 * support greater than the minimum requirement
 */
MINITEMS * newMinItemSetList( int size )
{
  MINITEMS * newMinSetList;
  int i;

  newMinSetList = ( MINITEMS * ) malloc ( sizeof ( MINITEMS ) );

  /*
   * I suppose for each i, there will be at least one itemset of
   * the size i , i ranges from 1 to the number of the items
   * so newMinSetList[i] contains the pointer to the list of all
   * item set with size i.
   */

  newMinSetList->listNum=size;
  newMinSetList->itemSetListOfSize
    = ( ITEMSETLIST * )malloc( size*sizeof ( ITEMSETLIST ) );
  newMinSetList->currentIdx=0;
  return newMinSetList;
}

void addItemSetList( MINITEMS * minSetList )
{
  int itNum, thisSize, nextSize;

  if ( minSetList->currentIdx>=minSetList->listNum ) {
    printf("number of itemset list is greater than expected \n ");
    exit( 1 );
  }

  itNum=( ( ( minSetList->itemSetListOfSize)[minSetList->currentIdx] ).itemNum )+1;
  /*
   * assuming that if the size of previous item set is n
   * then the size of this item set is n*(n-1)/2
   * maximum expectation
   */
  thisSize=( minSetList[ minSetList->currentIdx ] ).currentIdx;
  nextSize=thisSize*(thisSize-1)/2;
  
  newItemSetList( &( 
    ( minSetList->itemSetListOfSize )[minSetList->currentIdx] ), itNum, nextSize );
  minSetList->currentIdx++;
}

/*
 * remove the min-itemset list
 */
void delMinItemSetList( MINITEMS * setList )
{
  ITEMSETLIST * setListIdx;

  setListIdx=( setList->itemSetListOfSize )
    +( setList->currentIdx );
  while( setListIdx-- != setList->itemSetListOfSize ) {
    delItemSetList( setListIdx );
  }
  free ( setList->itemSetListOfSize );
  setList->itemSetListOfSize = NULL;
}

MINITEMS * newMinItSetList( DBASE * db )
{
  MINITEMS * minSetp;
  int * sumOfFields;
  int i,j,k;
  ITEMSET *cp;

  minSetp=newMinItemSetList( db->numOfItems );

  sumOfFields=(int *) malloc ( db->numOfItems*sizeof( int ) );

  /* initialize the number of items in each itemset list
   * for i>0, minSetp[i] means the list of item set whose
   * number of items in the set is `i'
   * minSet[0] is the list of `original' item set, which means
   * the set of all items in database before generating 
   * the `real' item set with all sizes
   */
  for( i=0; i<db->numOfItems; i++ ) {
    ( ( minSetp->itemSetListOfSize)[i] ).itemNum=i;
  }

  ( ( minSetp->itemSetListOfSize )[0] ).listSize=db->numOfItems;
  /* newItemSetList( minSetp->itemSetListOfSize, 1, db->numOfItems); */

  initMining( db->table, db->numOfRecs, db->numOfItems, minSetp );

  for ( i=2; i<=db->numOfItems; i++ ) {
    updateRel( db->table, db->numOfItems, minSetp, i );
  }

  return minSetp;
}

/*
 * In this function, we initialize the base of item set 
 * which means put each item in a seperate itemset
 */
void initMining( int * table, int recNum, int itemNum, MINITEMS * minSetp )
{
  int i=0,j=0,k=0, tempSum=0;
  int * recLst;
  ITEMSETLIST * curList;
  ITEMSET * curSet;

  newItemSetList( minSetp->itemSetListOfSize, 1, itemNum );
  curList=minSetp->itemSetListOfSize;

  for(i=0; i<itemNum; i++ ) {
    tempSum=0;
    k=0;
    recLst=( int * ) malloc( recNum*sizeof( int ) );
    for( j=0; j<recNum; j++ ) {
      if( table[ j*itemNum+i ]>0 ) {
        recLst[ k++ ]=j;
        tempSum+=table[ j*itemNum+i ];
      }
    }
    if( tempSum>=MINISUPPORT ) {
      curSet=( curList->setList )+(curList->currentIdx);
      newItemSet( curSet , 1 );
      ( curSet->idSet )[ 0 ]=i;
/* printf("curSet->idSet[0]=%d\n",(curSet->idSet)[0]); */
      curSet->support = tempSum;
      curSet->recSet=recLst;
      curSet->recCount=k;
      (curList->currentIdx)++;
/* printf("curList->currentIdx=%d\n",curList->currentIdx); */
    } else {
      free( recLst );
      recLst=NULL;
    }
/* printf("tempSum=%d\n",tempSum); */
  }
  minSetp->currentIdx=1;
}

ITEMSETLIST * updateRel( int * table, int colNum, MINITEMS * minSetp, int t )
{
  int i, j, k, prevSize;
  ITEMSETLIST * curISetListp, * prevItemSetList ;

  curISetListp = ( minSetp->itemSetListOfSize )+( minSetp->currentIdx );
  prevItemSetList=curISetListp-1;
  prevSize=prevItemSetList->currentIdx;
  newItemSetList( curISetListp, t, prevSize*(prevSize-1)/2 );
  newItemSet( curISetListp->setList, t );
printf("prevSize=%d\n",prevSize);
printf("t=%d\n",t);

  for ( i=0; i<prevSize; i++ ) {
    for ( j=i+1; j<prevSize; j++ ) {

      if( combineItemSet( table, colNum,
        &( ( prevItemSetList->setList )[ i ] ),
        &( ( prevItemSetList->setList )[ j ] ),
        &( ( curISetListp->setList )[ curISetListp->currentIdx ] ) ) ) {
/* printf("t=%d\n",t); */
printf("curISetListp->currentIdx=%d\n",curISetListp->currentIdx);
/* printf("(curISetListp->setList[ ( ( curISetListp->setList )->currentIdx)-1 ]).recCount=%d",
 (curISetListp->setList[ ( ( curISetListp->setList )->currentIdx)-1 ]).recCount ); */
          /* (curISetListp->setList[ ( curISetListp->setList )->currentIdx ]).recCount=
            occurCount( table, colNum,
              & ( prevItemSetList->setList[ i ] ), 
              & ( curISetListp->setList[ ( curISetListp->setList )->currentIdx ] ) ); */
          if( ( (curISetListp->setList[ curISetListp->currentIdx ]).recCount )
            >= MINISUPPORT ) {
printf("increasing currISetListp to %d!\n", curISetListp->currentIdx); 
            ( curISetListp->currentIdx )++;
          } else {
            delItemSet( (curISetListp->setList)+( ( curISetListp->setList )->currentIdx) );
          }
      }
    }
  }
  ( minSetp->currentIdx )++;
}

void showRelat( MINITEMS * minList )
{
  int i,j,k ;
  ITEMSETLIST * curItmLst;
  ITEMSET * curItm;

  for( i=0; i<minList->currentIdx; i++ ) {
    printf( "item set with %d elements \n", i+1 );
    curItmLst=( minList->itemSetListOfSize )+i;
    for( j=0; j<curItmLst->currentIdx; j++ ) {
      curItm=(curItmLst->setList)+j;
      printf("(");
      for( k=0; k<curItm->size; k++ ) {
        printf( "%d,",curItm->idSet[k] );
      }
      printf(")\t");
    }
    printf("\n");
  }
}
