#ifndef _DB_H_
#define _DB_H_
#include <stdio.h>

/* using bit vector to express the item list might
 * be better in the following definition in some
 * situation. However, the length of which is not
 * flexible, and it cannot record the number of items
 * in each transaction. So I decided to use an int
 * to represent that 
 * 
 * this struct is used to define a record in a
 * database. transId represents the transaction id,
 * and itemList, is the list of items in the transaction
 */ 
struct record {
  int transId;
  int * itemList;
};

typedef struct record RECORD;

/* this struct defines a table, whose length is the number of
 * records in this table. And its width defines the number of
 * fields in a table/database
 *
 * Here, the length and the width of the table are both dynamically
 * determined.
 */

struct database {
  int numOfRecs; /* record number */
  int numOfItems; /* field number */
  RECORD ** table; /* list of records */
};

typedef struct database DBASE;

/* allocate memory for a record or for a databse */
RECORD * newRec( int );
DBASE * newDb( int, int );

/* read from file */
DBASE * readFrom( FILE * );
void report( DBASE * );

/* remove record or database from memory */
/* void delRec( RECORD * ); */
void delDb( DBASE * );

/* insert a record in database at specified offset */
/* void insertRecAt( DBASE *, int, RECORD * ); */

/* finding out the sum of a field */
int sumOfField( DBASE * , int , int );

#define DATAITEM(tbl,i,j) ((((tbl)[i])->itemList)[j])

#endif
