#ifndef _ITEMSET_H_
#define _ITEMSET_H_

#include <string.h>
#include "mpi.h"
#include "db.h"

/* define the minimum confidence */
#define MINICONFID 6

/* define the minimum support */
#define MINISUPPORT 6

extern MPI_Status status;

/* in this class, the `size' is the number of items occur together
 * in a transaction. `idSet' is a list of ids of the items involved in
 * the itemset. Ids in the set are ordered
 * `support' is the data indicating how often they occure together
 */
struct itemSet {
  int size, currentIdx ;
  int support; 
  int recCount;/* the number of records having all the items in itemset */
  int * recSet;/* the record id set of itemset */
  int * idSet; /* can be considered as "int idSet[size]" */
  /* RECORD ** recsInItemSet; */
  /* can be considered as "(RECORD *) recsInItemSet[size] */
};

typedef struct itemSet ITEMSET;

struct itemSetLinkList {
  int * idSet;
  int occurCount, itemCount;
  struct itemSetLinkList * next;
};

typedef struct itemSetLinkList ITLSTNODE;

/*
 * In this class, the `listSize' is the number of items sets in this
 * list. `itemNum' is the size of each itemset (itemSet.size ). 
 * this structure contains an array of itemSet, 
 */
struct itemSetList {
  int itemNum, listSize, currentIdx; 
  ITEMSET * setList; 
  int * recCountLst;
};

typedef struct itemSetList ITEMSETLIST;

/* an array of pointers */
struct minItemSetList {
  int listNum, currentIdx;
  ITEMSETLIST * itemSetListOfSize;
  /* ITEMSETLIST itemSetListOfAllSize[listSize]; */
};

typedef struct minItemSetList MINITEMS;

void newItemSet( ITEMSET * , int );
void addAnItem( ITEMSET * , int );
int equalItemSet( ITEMSET * , ITEMSET * );
int combineItemSet( int * , int , ITEMSET * , ITEMSET * , ITEMSET * );
void delItemSet( ITEMSET * );
void newItemSetList( ITEMSETLIST * , int , int );
void addAnItemSet( ITEMSETLIST * );
void delItemSetList( ITEMSETLIST * );
MINITEMS * newMinItemSetList( int );
void addItemSetList( MINITEMS * );
void delMinItemSetList( MINITEMS * );
ITLSTNODE * newMinItSetList( int * , int , int );
void initMining( int * , int , int , MINITEMS * );
ITEMSETLIST * updateRel( int *, int, MINITEMS * , int );
int occurCount( int * , int , ITEMSET * , ITEMSET * );
void showRelat( MINITEMS * );
int sortMinLst( ITLSTNODE *, int );
void copyIds( ITLSTNODE * , int * , int );
void itemSetLocalUpdt( int *, MINITEMS *, int, int );
void copyItemSet( ITEMSET * , ITEMSET * );

#define LOCALBUFSIZE 1024

#endif
