#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>


#define maxn 4

#define nxp 2
#define nyp 2

#define nx 4
#define ny 4

#define ndim 2

main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, value, size, i, j;
    int up_nbr, down_nbr;
    int my_grid_rank;
    
    double toterr=0.0, myerr=0.0;

    MPI_Comm   grid_comm;
    MPI_Datatype  rowtype;
    
    int coord[ndim];
    int dims[2];
    int periods[2];
        
    int nxblock=nx /nxp;
    int nyblock=ny /nyp;
    int nxlocal=nx/nxp +1;
    int nylocal=ny/nyp +1;
    int reorder;
    int coordinates[2];

    int nbrleft, nbrright, nbrbottom, nbrtop;
    
    double u[maxn][maxn];
    double unew[maxn][maxn];
    double f[maxn][maxn]; 
    double dx=1.0/nx;
    double dy=1.0/ny;
    
    dims[0] = nxp;
    dims[1] = nyp;
    
    /* no circular shift */
    periods[0] = 0;
    periods[1] = 0;
    
    reorder = 1;
    
    MPI_Init(&argc, &argv);


    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
		    &grid_comm);

    if (grid_comm != MPI_COMM_NULL)
      {
	MPI_Comm_rank(grid_comm, &my_grid_rank);
	MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);

	MPI_Cart_shift(grid_comm, 1, 1, &nbrleft, &nbrright);
	MPI_Cart_shift(grid_comm, 0, 1, &nbrbottom, &nbrtop);

	MPI_Type_vector(nyblock, 1, nxlocal+1, MPI_DOUBLE, &rowtype);
	MPI_Type_commit(&rowtype);
    
	for (j=0; nylocal; j++) 
	  {
	    for (i=0; nxlocal; i++) 
	      {
		f[i][j] = - 2.0 *((dx *(i + coordinates[2] * nxblock)) *
				  (dx *(i + coordinates[2] * nxblock)))
		  + 2.0 * dx * (i + coordinates[2] * nxblock);
		u[i][j] = 0.0;
		unew[i][j] = 0.0;
	      }
	  }
      }

    /*    do {  */
      if (grid_comm != MPI_COMM_NULL) {
	MPI_Sendrecv( &u[1][1], nxblock, MPI_DOUBLE, nbrleft, 0, 
		      &u[1][nylocal], nxblock, MPI_DOUBLE, nbrright, 0, 
		      grid_comm, &status );

	/*  nylocal  --->  nylocal-1  */
	MPI_Sendrecv( &u[1][nylocal], nxblock, MPI_DOUBLE, nbrright, 1,
		      &u[1][0], nxblock, MPI_DOUBLE, nbrleft, 1, 
		      grid_comm, &status );

	MPI_Sendrecv( &u[1][1], 1, rowtype, nbrtop, 2, 
		      &u[nxlocal][1], 1, rowtype, nbrbottom, 2, 
		      grid_comm, &status );

	MPI_Sendrecv( &u[nxlocal-1][1], 1, rowtype, nbrbottom, 3,
		      &u[0][1], 1, rowtype, nbrtop, 3, 
		      grid_comm, &status );
      }
    
    /* Check that we have the correct results */
      myerr= 0.0;
      for (j=1; nylocal-1; j++)
	{
	  for (i=1; nxlocal-1; i++)
	    {
	      unew[i][j] = (u[i-1][j] + u[i+1][j] +
			    u[i][j-1] + u[i][j+1] +
			    f[i][j])/4.0;
	      if (fabs( unew[i][j] - u[i][j]) > toterr)
		myerr = fabs(unew[i][j] - u[i][j]);
	      else
		myerr = toterr;
	    }
	}

      MPI_Reduce( &myerr, &toterr, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

      for (j=1; nylocal-1; j++)
	{
	  for (i=1; nxlocal-1; i++)
	    {
	      u[i][j]=unew[i][j];
	    }
	}
      /*    } while(toterr > 1.0e-3);  */
    
    if (my_grid_rank == 0) {
	if (toterr)
	    printf( "! found %lf errors\n", toterr );
	else
	    printf( "No errors\n" );
    }
    MPI_Finalize( );
}

