#include <stdio.h>
#include <mpi.h>
#include <math.h>

void main(int argc, char *argv[]) {

/* number of interval = 64
   buffer size = 8, since there are 8 processors available
*/
  int myid, numprocs, i, n = 64;
  double PI25DT = 3.14159265358979323842643;
  double mypi, pi, h, sum, x, buf[8];

  MPI_Init(&argc, &argv);

  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);

  MPI_Comm_rank(MPI_COMM_WORLD, &myid);

  h = 1.0 /(double) n;
  sum = 0.0;

/* computes local pi value */

  for (i=myid+1; i<=n; i+=numprocs) {
    x = h * ((double)i - 0.5);
    sum += 4.0 / (1.0 + x*x);
  }
  mypi = h * sum;

/* send local pi to the root process */

  MPI_Gather(&mypi, 1, MPI_DOUBLE, &buf, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  if (myid == 0) {

/* computs global pi */

    for(i=0, pi=0; i<numprocs; i++)
      pi += buf[i];

    printf("pi is approximately %.16f, and the Error is %.16f\n",
            pi, fabs(pi - PI25DT));
  }

/* broadcast the global pi and print */

  MPI_Bcast(&pi, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  printf("process:%d, pi:%f\n", myid, pi);

  MPI_Finalize();
}
