#include <mpi.h>
#include <stdio.h>
#include <math.h>

void main(int argc, char **argv) {

  double pisquared, sum, h, x, y;
  double PI25DT = 3.14159265358979323842643;
  int pid, size, i, j, grid_rank, n = 5000;
  int dim[2], wrap[2], coords[2];
  MPI_Comm comm_grid;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &pid);
  dim[0] = 4;                               /* 2 x 4 processors */
  dim[1] = 2;
  wrap[0] = wrap[1] = 0;
  MPI_Cart_create(MPI_COMM_WORLD, 2, dim, wrap, 0, &comm_grid);
  MPI_Comm_rank(comm_grid, &grid_rank);
  MPI_Cart_coords(comm_grid, grid_rank, 2, coords);

  sum = 0;
  h = 1/(double)n;

  for (i=coords[0]; i<n; i+=2) {
    for (j=coords[1]; j<n; j+=4)  {
      x = (double)i * h;
      y = (double)j * h;
      sum += h*h*(1/((1+(x+.5*h)*(x+.5*h))*(1+(y+.5*h)*(y+.5*h))));
    }
  }
  
  sum *= 16;
  MPI_Reduce(&sum, &pisquared, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if (pid == 0)
    printf("pi square is:%0.16f, and error is:%0.16f\n",
	   pisquared, fabs(pisquared - PI25DT * PI25DT));
  MPI_Finalize();
}
